# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Query LLM for structured output."""

import contextlib
import dataclasses
import functools
import inspect
import time
from typing import Annotated, Any, Callable, ClassVar, Iterator, Type, Union
import uuid

import langfun.core as lf
from langfun.core.structured import mapping
from langfun.core.structured import schema as schema_lib
import pyglove as pg


@lf.use_init_args(['schema', 'default', 'examples'])
class LfQuery(mapping.Mapping):
  """Base class for different implementations of `lf.query`.

  By subclassing this class, users could create variations of prompts for
  `lf.query` and associated them with specific protocols and versions.

  For example:

  ```
  class _MyLfQuery(LFQuery):
    protocol = 'my_format'
    version = '1.0'

    template_str = inspect.cleandoc(
      '''
      ...
      '''
    )
    mapping_template = lf.Template(
      '''
      ...
      '''
    )

  lf.query(..., protocol='my_format:1.0')
  ```

  (THIS IS NOT A TEMPLATE)
  """

  context_title = 'CONTEXT'
  input_title = 'INPUT_OBJECT'

  # Mark schema as required.
  schema: pg.typing.Annotated[
      schema_lib.schema_spec(), 'Required schema for parsing.'
  ]

  # A map from (protocol, version) to the query structure class.
  # This is used to map different protocols/versions to different templates.
  # So users can use `lf.query(..., protocol='<protocol>:<version>')` to use
  # a specific version of the prompt. We use this feature to support variations
  # of prompts and maintain backward compatibility.
  _OOP_PROMPT_MAP: ClassVar[
      dict[
          str,        # protocol.
          dict[
              str,    # version.
              Type['LfQuery']
          ]
      ]
  ] = {}

  # This the flag to update default protocol version.
  _DEFAULT_PROTOCOL_VERSIONS: ClassVar[dict[str, str]] = {
      'python': '2.0',
      'json': '1.0',
  }

  def __init_subclass__(cls) -> Any:
    super().__init_subclass__()
    if not inspect.isabstract(cls):
      protocol = cls.__schema__['protocol'].default_value
      version_dict = cls._OOP_PROMPT_MAP.get(protocol)
      if version_dict is None:
        version_dict = {}
        cls._OOP_PROMPT_MAP[protocol] = version_dict
      dest_cls = version_dict.get(cls.version)
      if dest_cls is not None and dest_cls.__type_name__ != cls.__type_name__:
        raise ValueError(
            f'Version {cls.version} is already registered for {dest_cls!r} '
            f'under protocol {protocol!r}. Please use a different version.'
        )
      version_dict[cls.version] = cls

  @classmethod
  def from_protocol(cls, protocol: str) -> Type['LfQuery']:
    """Returns a query structure from the given protocol and version."""
    if ':' in protocol:
      protocol, version = protocol.split(':')
    else:
      version = cls._DEFAULT_PROTOCOL_VERSIONS.get(protocol)
      if version is None:
        version_dict = cls._OOP_PROMPT_MAP.get(protocol)
        if version_dict is None:
          raise ValueError(
              f'Protocol {protocol!r} is not supported. Available protocols: '
              f'{sorted(cls._OOP_PROMPT_MAP.keys())}.'
          )
        elif len(version_dict) == 1:
          version = list(version_dict.keys())[0]
        else:
          raise ValueError(
              f'Multiple versions found for protocol {protocol!r}, please '
              f'specify a version with "{protocol}:<version>".'
          )

    version_dict = cls._OOP_PROMPT_MAP.get(protocol)
    if version_dict is None:
      raise ValueError(
          f'Protocol {protocol!r} is not supported. Available protocols: '
          f'{sorted(cls._OOP_PROMPT_MAP.keys())}.'
      )
    dest_cls = version_dict.get(version)
    if dest_cls is None:
      raise ValueError(
          f'Version {version!r} is not supported for protocol {protocol!r}. '
          f'Available versions: {sorted(version_dict.keys())}.'
      )
    return dest_cls


class _LfQueryJsonV1(LfQuery):
  """Query a structured value using JSON as the protocol."""

  preamble = """
      Please respond to the last {{ input_title }} with {{ output_title}} according to {{ schema_title }}:

      INSTRUCTIONS:
        1. If the schema has `_type`, carry it over to the JSON output.
        2. If a field from the schema cannot be extracted from the response, use null as the JSON value.

      {{ input_title }}:
        1 + 1 =

      {{ schema_title }}:
        {"result": {"_type": "langfun.core.structured.query.Answer", "final_answer": int}}

      {{ output_title}}:
        {"result": {"_type": "langfun.core.structured.query.Answer", "final_answer": 2}}
      """

  version = '1.0'
  protocol = 'json'
  schema_title = 'SCHEMA'
  output_title = 'JSON'


class _LfQueryPythonV1(LfQuery):
  """Query a structured value using Python as the protocol."""

  preamble = """
      Please respond to the last {{ input_title }} with {{ output_title }} according to {{ schema_title }}.

      {{ input_title }}:
        1 + 1 =

      {{ schema_title }}:
        Answer

        ```python
        class Answer:
          final_answer: int
        ```

      {{ output_title }}:
        ```python
        Answer(
          final_answer=2
        )
        ```
      """
  version = '1.0'
  protocol = 'python'
  schema_title = 'OUTPUT_TYPE'
  output_title = 'OUTPUT_OBJECT'
  mapping_template = lf.Template(
      """
      {%- if example.context -%}
      {{ context_title}}:
      {{ example.context | indent(2, True)}}

      {% endif -%}

      {{ input_title }}:
      {{ example.input_repr(protocol, compact=False) | indent(2, True) }}

      {% if example.schema -%}
      {{ schema_title }}:
      {{ example.schema_repr(protocol) | indent(2, True) }}

      {% endif -%}

      {{ output_title }}:
      {%- if example.has_output %}
      {{ example.output_repr(protocol, compact=False) | indent(2, True) }}
      {% endif -%}
      """
  )


class _LfQueryPythonV2(LfQuery):
  """Query a structured value using Python as the protocol."""

  preamble = """
      Please respond to the last {{ input_title }} with {{ output_title }} only according to {{ schema_title }}.

      {{ input_title }}:
        1 + 1 =

      {{ schema_title }}:
        Answer

        ```python
        class Answer:
          final_answer: int
        ```

      {{ output_title }}:
        ```python
        output = Answer(
          final_answer=2
        )
        ```
      """
  version = '2.0'
  protocol = 'python'
  input_title = 'REQUEST'
  schema_title = 'OUTPUT PYTHON TYPE'
  output_title = 'OUTPUT PYTHON OBJECT'
  mapping_template = lf.Template(
      """
      {%- if example.context -%}
      {{ context_title}}:
      {{ example.context | indent(2, True)}}

      {% endif -%}

      {{ input_title }}:
      {{ example.input_repr(protocol, compact=False) | indent(2, True) }}

      {% if example.schema -%}
      {{ schema_title }}:
      {{ example.schema_repr(protocol) | indent(2, True) }}

      {% endif -%}

      {{ output_title }}:
      {%- if example.has_output %}
      {{ example.output_repr(protocol, compact=False, assign_to_var='output') | indent(2, True) }}
      {% endif -%}
      """
  )


def query(
    prompt: Union[str, lf.Template, Any],
    schema: schema_lib.SchemaType |  None = None,
    default: Any = lf.RAISE_IF_HAS_ERROR,
    *,
    lm: lf.LanguageModel | list[lf.LanguageModel],
    num_samples: int | list[int] = 1,
    system_message: str | lf.Template | None = None,
    examples: list[mapping.MappingExample] | None = None,
    cache_seed: int | None = 0,
    response_postprocess: Callable[[str], str] | None = None,
    autofix: int = 0,
    autofix_lm: lf.LanguageModel | None = None,
    protocol: str | None = None,
    returns_message: bool = False,
    skip_lm: bool = False,
    invocation_id: str | None = None,
    **kwargs,
) -> Any:
  """Query one or more language models for structured or unstructured outputs.

  This is the primary API in Langfun for interacting with language models,
  supporting natural language prompts, structured inputs, and multiple advanced
  features.

  Key Features:

    - **Input**: Accepts natural language strings, structured inputs (e.g.,
      `pg.Object`), and templates (`lf.Template`) with modality objects.

    - **Output**: Returns structured outputs when `schema` is specified;
      otherwise, outputs raw natural language (as a string).

    - **Few-shot examples**: Supports structured few-shot examples with the
      `examples` argument.

    - **Multi-LM fan-out**: Sends queries to multiple language models with in
      multiple samples in parallel,  returning a list of outputs.

  Examples:

    Case 1: Regular natural language-based LLM query:

    ```
    lf.query('1 + 1 = ?', lm=lf.llms.Gpt4Turbo())

    # Outptut: '2'
    ```

    Case 2: Query with structured output.

    ```
    lf.query('1 + 1 = ?', int, lm=lf.llms.Gpt4Turbo())

    # Output: 2
    ```

    Case 3: Query with structured input.

    ```
    class Sum(pg.Object):
      a: int
      b: int

    lf.query(Sum(1, 1), int, lm=lf.llms.Gpt4Turbo())

    # Output: 2
    ```

    Case 4: Query with input of mixed modalities.

    ```
    class Animal(pg.Object):
      pass

    class Dog(Animal):
      pass

    class Entity(pg.Object):
      name: str

    lf.query(
        'What is in this {{image}} and {{objects}}?'
        list[Entity],
        lm=lf.llms.Gpt4Turbo()
        image=lf.Image(path='/path/to/a/airplane.png'),
        objects=[Dog()],
    )

    # Output: [Entity(name='airplane'), Entity(name='dog')]
    ```

    Case 5: Query with structured few-shot examples.
    ```
    lf.query(
        'What is in this {{image}} and {{objects}}?'
        list[Entity],
        lm=lf.llms.Gpt4Turbo()
        image=lf.Image(path='/path/to/a/dinasaur.png'),
        objects=[Dog()],
        examples=[
            lf.MappingExample(
                input=lf.Template(
                    'What is the object near the house in this {{image}}?',
                    image=lf.Image(path='/path/to/image.png'),
                ),
                schema=Entity,
                output=Entity('cat'),
            ),
        ],
    )

    # Output: [Entity(name='dinasaur'), Entity(name='dog')]
    ```

    Case 6: Multiple queries to multiple models.
    ```
    lf.query(
        '1 + 1 = ?',
        int,
        lm=[
            lf.llms.Gpt4Turbo(),
            lf.llms.Gemini1_5Pro(),
        ],
        num_samples=[1, 2],
    )
    # Output: [2, 2, 2]
    ```

  Args:
    prompt: The input query. Can be:
      - A natural language string (supports templating with `{{}}`),
      - A `pg.Object` object for structured input,
      - An `lf.Template` for mixed or template-based inputs.
    schema: Type annotation or `lf.Schema` object for the expected output. 
      If `None` (default), the response will be a natural language string.
    default: Default value to return if parsing fails. If not specified, an
      error will be raised.
    lm: The language model(s) to query. Can be:
      - A single `LanguageModel`,
      - A list of `LanguageModel`s for multi-model fan-out.
    num_samples: Number of samples to generate. If a list is provided, its
      length must match the number of models in `lm`.
    system_message: System instructions to guide the model output. If None,
      no system message will be used.
    examples: Few-shot examples to guide the model output. Defaults to `None`.
    cache_seed: Seed for caching the query. Queries with the same
      `(lm, prompt, cache_seed)` will use cached responses. If `None`,
      caching is disabled.
    response_postprocess: A post-processing function for the raw LM response.
      If `None`, no post-processing occurs.
    autofix: Number of attempts for auto-fixing code errors. Set to `0` to
      disable auto-fixing. Not supported with the `'json'` protocol.
    autofix_lm: The LM to use for auto-fixing. Defaults to the `autofix_lm`
      from `lf.context` or the main `lm`.
    protocol: Format for schema representation. Builtin choices are `'json'` or
      `'python'`, users could extend with their own protocols by subclassing
      `lf.structured.LfQuery'. Also protocol could be specified with a version
      in the format of 'protocol:version', e.g., 'python:1.0', so users could
      use a specific version of the prompt based on the protocol. Please see the
      documentation of `LfQuery` for more details. If None, the protocol from
      context manager `lf.query_protocol` will be used, or 'python' if not
      specified.
    returns_message:  If `True`, returns an `lf.Message` object instead of
      the final parsed result.
    skip_lm: If `True`, skips the LLM call and returns the rendered 
      prompt as a `UserMessage` object.
    invocation_id: The ID of the query invocation, which will be passed to
      `lf.QueryInvocation` when `lf.trackIf `None`, a unique ID will
      be generated.
    **kwargs: Additional keyword arguments for:
      - Rendering templates (e.g., `template_str`, `preamble`),
      - Configuring `lf.structured.Mapping`.

  Returns:
    The result of the query:
    - A single output or a list of outputs if multiple models/samples are used.
    - Each output is a parsed object matching `schema`, an `lf.Message` (if 
      `returns_message=True`), or a natural language string (default).
  """
    # Internal usage logging.

  if protocol is None:
    protocol = lf.context_value('__query_protocol__', 'python')

  def _invocation_id():
    return invocation_id or f'query@{uuid.uuid4().hex[-7:]}'

  # Multiple quries will be issued when `lm` is a list or `num_samples` is
  # greater than 1.
  if isinstance(lm, list) or num_samples != 1:
    def _single_query(inputs):
      i, (lm, example_i) = inputs
      return query(
          prompt,
          schema,
          default=default,
          lm=lm,
          system_message=system_message,
          examples=examples,
          # Usually num_examples should not be large, so we multiple the user
          # provided cache seed by 100 to avoid collision.
          cache_seed=(
              None if cache_seed is None else cache_seed * 100 + example_i
          ),
          response_postprocess=response_postprocess,
          autofix=autofix,
          autofix_lm=autofix_lm,
          protocol=protocol,
          returns_message=returns_message,
          skip_lm=skip_lm,
          invocation_id=f'{_invocation_id()}:{i}',
          **kwargs,
      )
    lm_list = lm if isinstance(lm, list) else [lm]
    num_samples_list = (
        num_samples if isinstance(num_samples, list)
        else [num_samples] * len(lm_list)
    )
    assert len(lm_list) == len(num_samples_list), (
        'Expect the length of `num_samples` to be the same as the '
        f'the length of `lm`. Got {num_samples} and {lm_list}.'
    )
    query_inputs = []
    total_queries = 0
    for lm, num_samples in zip(lm_list, num_samples_list):
      query_inputs.extend([(lm, i) for i in range(num_samples)])
      total_queries += num_samples

    samples = []
    for _, output, error in lf.concurrent_map(
        _single_query, enumerate(query_inputs),
        max_workers=max(64, total_queries),
        ordered=True,
    ):
      if error is None:
        samples.append(output)
    return samples

  # Normalize query schema.
  # When `lf.query` is used for symbolic completion, schema is automatically
  # inferred when it is None.
  if isinstance(prompt, pg.Symbolic) and prompt.sym_partial and schema is None:
    schema = prompt.__class__

  # Attach system message as input template metadata, which will be passed
  # through to the rendered message metadata under key `system_message`.
  if system_message is not None:
    kwargs['metadata_system_message'] = lf.Template.from_value(
        system_message
    ).render(message_cls=lf.SystemMessage)

  # Normalize query input.
  if isinstance(prompt, (lf.Message, str)):
    # Query with structured output.
    prompt_kwargs = kwargs.copy()
    prompt_kwargs.pop('template_str', None)
    query_input = lf.Template.from_value(prompt, **prompt_kwargs)
  elif isinstance(prompt, lf.Template):
    # Create a copy of the prompt if it has a parent object, so all child
    # modality objects could be referred by path relative to the prompt.
    query_input = prompt.clone() if prompt.sym_parent is not None else prompt

    # Attach template metadata from kwargs. This is used to pass through fields
    # from kwargs to the rendered message.
    template_metadata = {
        k: v for k, v in kwargs.items() if k.startswith('metadata_')
    }
    query_input.rebind(
        template_metadata, skip_notification=True, raise_on_no_change=False
    )
  elif pg.MISSING_VALUE == prompt:
    query_input = lf.UserMessage('')
  else:
    query_input = schema_lib.mark_missing(prompt)

  # Determine query class.
  if schema in (None, str):
    # Non-structured query.
    query_cls = None
  else:
    # Query with structured output.
    query_cls = LfQuery.from_protocol(protocol)
    if ':' not in protocol:
      protocol = f'{protocol}:{query_cls.version}'

  # `skip_lm`` is True when `lf.query_prompt` is called.
  # and `prompt` is `pg.MISSING_VALUE` when `lf.query_output` is called.
  # In these cases, we do not track the query invocation.
  if skip_lm or pg.MISSING_VALUE == prompt:
    trackers = []
  else:
    trackers = lf.context_value('__query_trackers__', [])

  # Mark query start with trackers.
  # NOTE: prompt is MISSING_VALUE when `lf.query_output` is called.
  # We do not track the query invocation in this case.
  if trackers:
    invocation = QueryInvocation(
        id=_invocation_id(),
        input=pg.Ref(query_input),
        schema=(
            schema_lib.Schema.from_value(schema)
            if schema not in (None, str) else None
        ),
        default=default,
        lm=pg.Ref(lm),
        examples=pg.Ref(examples) if examples else [],
        protocol=protocol,
        kwargs={k: pg.Ref(v) for k, v in kwargs.items()},
        start_time=time.time(),
    )
    for i, tracker in enumerate(trackers):
      if i == 0 or tracker.include_child_scopes:
        tracker.track(invocation)
  else:
    invocation = None

  def _mark_query_completed(output_message, error, usage_summary):
    # Mark query completion with trackers.
    if not trackers:
      return

    if output_message is not None:
      # To minimize payload for serialization, we remove the result and usage
      # fields from the metadata. They will be computed on the fly when the
      # invocation is rendered.
      metadata = dict(output_message.metadata)
      metadata.pop('result', None)
      metadata.pop('usage', None)
      lm_response = lf.AIMessage(output_message.text, metadata=metadata)
    else:
      lm_response = None

    assert invocation is not None
    invocation.mark_completed(
        lm_response=lm_response, error=error, usage_summary=usage_summary,
    )
    for i, tracker in enumerate(trackers):
      if i == 0 or tracker.include_child_scopes:
        tracker.mark_completed(invocation)

  with lf.track_usages() as usage_summary:
    try:
      if query_cls is None:
        # Query with natural language output.
        output_message = lf.LangFunc.from_value(query_input, **kwargs)(
            lm=lm, cache_seed=cache_seed, skip_lm=skip_lm
        )
        if response_postprocess:
          processed_text = response_postprocess(output_message.text)
          if processed_text != output_message.text:
            output_message = lf.AIMessage(processed_text, source=output_message)
      else:
        # Query with structured output.
        output_message = query_cls(
            input=(
                query_input.render(lm=lm)
                if isinstance(query_input, lf.Template)
                else query_input
            ),
            schema=schema,
            examples=examples,
            response_postprocess=response_postprocess,
            autofix=autofix if protocol.startswith('python:') else 0,
            **kwargs,
        )(
            lm=lm,
            autofix_lm=autofix_lm or lm,
            cache_seed=cache_seed,
            skip_lm=skip_lm,
        )
      _mark_query_completed(output_message, None, usage_summary)
    except mapping.MappingError as e:
      _mark_query_completed(
          e.lm_response, pg.ErrorInfo.from_exception(e), usage_summary
      )
      if lf.RAISE_IF_HAS_ERROR == default:
        raise e
      output_message = e.lm_response
      output_message.result = default
    except BaseException as e:
      _mark_query_completed(
          None, pg.ErrorInfo.from_exception(e), usage_summary
      )
      raise e

  if returns_message:
    return output_message
  return output_message.text if schema in (None, str) else output_message.result


async def aquery(
    prompt: Union[str, lf.Template, Any],
    schema: schema_lib.SchemaType |  None = None,
    default: Any = lf.RAISE_IF_HAS_ERROR,
    *,
    lm: lf.LanguageModel | list[lf.LanguageModel] | None = None,
    num_samples: int | list[int] = 1,
    system_message: str | lf.Template | None = None,
    examples: list[mapping.MappingExample] | None = None,
    cache_seed: int | None = 0,
    response_postprocess: Callable[[str], str] | None = None,
    autofix: int = 0,
    autofix_lm: lf.LanguageModel | None = None,
    protocol: str | None = None,
    returns_message: bool = False,
    skip_lm: bool = False,
    invocation_id: str | None = None,
    **kwargs,
) -> Any:
  """Async version of `lf.query`."""
  # TODO(daiyip): implement native async querying.
  return await lf.invoke_async(
      query,
      prompt,
      schema,
      default,
      lm=lm,
      num_samples=num_samples,
      system_message=system_message,
      examples=examples,
      cache_seed=cache_seed,
      response_postprocess=response_postprocess,
      autofix=autofix,
      autofix_lm=autofix_lm,
      protocol=protocol,
      returns_message=returns_message,
      skip_lm=skip_lm,
      invocation_id=invocation_id,
      **kwargs
  )


@contextlib.contextmanager
def query_protocol(protocol: str) -> Iterator[None]:
  """Context manager for setting the query protocol for the scope."""
  with lf.context(__query_protocol__=protocol):
    try:
      yield
    finally:
      pass

#
# Helper function for map-reduce style querying.
#


def query_and_reduce(
    prompt: Union[str, lf.Template, Any],
    schema: schema_lib.SchemaType | None = None,
    *,
    reduce: Callable[[list[Any]], Any],
    lm: lf.LanguageModel | list[lf.LanguageModel] | None = None,
    num_samples: int | list[int] = 1,
    **kwargs,
) -> Any:
  """Issues multiple `lf.query` calls in parallel and reduce the outputs.
  
  Args:
    prompt: A str (may contain {{}} as template) as natural language input, or a
      `pg.Symbolic` object as structured input as prompt to LLM.
    schema: A type annotation as the schema for output object. If str (default),
      the response will be a str in natural language.
    reduce: A function to reduce the outputs of multiple `lf.query` calls. It
      takes a list of outputs and returns the final object.
    lm: The language model to use. If not specified, the language model from
      `lf.context` context manager will be used.
    num_samples: The number of samples to obtain from each language model being
      requested. If a list is provided, it should have the same length as `lm`.
    **kwargs: Additional arguments to pass to `lf.query`.

  Returns:
    The reduced output from multiple `lf.query` calls.
  """
  results = query(prompt, schema, lm=lm, num_samples=num_samples, **kwargs)
  if isinstance(results, list):
    results = reduce(results)
  return results


#
# Functions for decomposing `lf.query` into pre-llm and post-llm operations.
#


def query_prompt(
    prompt: Union[str, lf.Template, Any],
    schema: schema_lib.SchemaType | None = None,
    **kwargs,
) -> lf.Message:
  """Returns the final prompt sent to LLM for `lf.query`."""
  # Delay import to avoid circular dependency in Colab.
  # llms > data/conversion > structured > querying
  from langfun.core.llms import fake  # pylint: disable=g-import-not-at-top

  kwargs.pop('returns_message', None)
  kwargs.pop('skip_lm', None)
  return query(
      prompt, schema,
      # The LLM will never be used, it's just a placeholder.
      lm=fake.Pseudo(),
      skip_lm=True,
      returns_message=True,
      **kwargs
  )


def query_output(
    response: Union[str, lf.Message],
    schema: schema_lib.SchemaType | None = None,
    **kwargs,
) -> Any:
  """Returns the final output of `lf.query` from a provided LLM response."""
  # Delay import to avoid circular dependency in Colab.
  # llms > data/conversion > structured > querying
  from langfun.core.llms import fake  # pylint: disable=g-import-not-at-top

  kwargs.pop('prompt', None)
  kwargs.pop('lm', None)
  return query(
      pg.MISSING_VALUE, schema, lm=fake.StaticResponse(response), **kwargs
  )


#
# Functions for computing reward of an LLM response based on a mapping example.
#


def query_reward(
    mapping_example: Union[str, mapping.MappingExample],
    response: Union[str, lf.Message],
) -> float | None:
  """Returns the reward of an LLM response based on an mapping example."""
  if isinstance(mapping_example, str):
    mapping_example = pg.from_json_str(mapping_example)
    assert isinstance(mapping_example, mapping.MappingExample), mapping_example
  schema = mapping_example.schema

  if schema and isinstance(schema.spec, pg.typing.Object):
    output_cls = schema.spec.cls
  elif schema is None and isinstance(mapping_example.output, pg.Object):
    output_cls = mapping_example.output.__class__
  else:
    output_cls = None

  reward_fn = _reward_fn(output_cls)
  if reward_fn is None:
    return None

  return reward_fn(
      query_output(response, output_cls),
      mapping_example.input,
      mapping_example.output,
      mapping_example.metadata,
  )


@functools.cache
def _reward_fn(cls) -> Callable[
    [
        pg.Object,    # Actual output object.
        Any,          # Input object.
        pg.Object,    # Expected output object.
        pg.Dict       # User metadata.
    ], float] | None:
  """Returns the reward function for a class that is being queried."""
  if not callable(getattr(cls, '__reward__', None)):
    return None

  signature = pg.typing.signature(cls.__reward__)
  num_args = len(signature.args)
  if num_args < 2 or num_args > 4:
    raise TypeError(
        f'`{cls.__type_name__}.__reward__` should have signature: '
        '`__reward__(self, input, [expected_output], [expected_metadata])`.'
    )
  def _reward(self, input, expected_output, metadata):  # pylint: disable=redefined-builtin
    args = [self, input, expected_output, metadata]
    return cls.__reward__(*args[:num_args])
  return _reward


#
# Functions for tracking `lf.query` invocations.
#


class QueryInvocation(pg.Object, pg.views.HtmlTreeView.Extension):
  """A class to represent the invocation of `lf.query`."""

  #
  # Query input.
  #

  id: Annotated[
      str,
      'The ID of the query invocation.'
  ]

  input: Annotated[
      Union[lf.Template, pg.Symbolic],
      'Mapping input of `lf.query`.'
  ]

  schema: pg.typing.Annotated[
      schema_lib.schema_spec(noneable=True),
      'Schema of `lf.query`.'
  ]

  default: Annotated[
      Any,
      'Default value of `lf.query`.'
  ] = lf.RAISE_IF_HAS_ERROR

  lm: Annotated[
      lf.LanguageModel,
      'Language model used for `lf.query`.'
  ]

  examples: Annotated[
      list[mapping.MappingExample],
      'Fewshot exemplars for `lf.query`.'
  ]

  protocol: Annotated[
      str,
      'Protocol of `lf.query`.'
  ] = 'python'

  kwargs: Annotated[
      dict[str, Any],
      'Kwargs of `lf.query`.'
  ] = {}

  #
  # Query output.
  #

  lm_response: Annotated[
      lf.Message | None,
      'Raw LM response. If None, query is not completed yet or failed.'
  ] = None

  error: Annotated[
      pg.ErrorInfo | None,
      'Error info if the query failed.'
  ] = None

  #
  # Execution details.
  #

  start_time: Annotated[
      float,
      'Start time of query.'
  ]

  end_time: Annotated[
      float | None,
      'End time of query. If None, query is not completed yet.'
  ] = None

  usage_summary: Annotated[
      lf.UsageSummary,
      'Usage summary of the query.'
  ] = lf.UsageSummary()

  @functools.cached_property
  def lm_request(self) -> lf.Message:
    return query_prompt(
        self.input, self.schema, examples=self.examples or None,
        protocol=self.protocol,
        **self.kwargs
    )

  @property
  def output(self) -> Any:
    """The output of `lf.query`. If it failed, returns None."""
    return self._output

  @property
  def has_error(self) -> bool:
    """Returns True if the query failed to generate a valid output."""
    return self.error is not None

  @property
  def has_oop_error(self) -> bool:
    """Returns True if the query failed due to out of memory error."""
    return self.error is not None and self.error.tag.startswith('MappingError')

  @property
  def elapse(self) -> float:
    """Returns query elapse in seconds."""
    if self.end_time is None:
      return time.time() - self.start_time
    return self.end_time - self.start_time

  def as_mapping_example(
      self,
      metadata: dict[str, Any] | None = None
  ) -> mapping.MappingExample:
    """Returns a `MappingExample` object for this query invocation."""
    return mapping.MappingExample(
        input=self.input,
        schema=self.schema,
        output=self.lm_response.text if self.has_oop_error else self.output,
        metadata=metadata or {},
    )

  def _on_bound(self):
    super()._on_bound()
    self._tab_control = None
    self._output = None
    self.__dict__.pop('lm_request', None)

  @property
  def is_completed(self) -> bool:
    """Returns True if the query is completed."""
    return self.end_time is not None

  def mark_completed(
      self,
      lm_response: lf.Message | None,
      error: pg.ErrorInfo | None = None,
      usage_summary: lf.UsageSummary | None = None) -> None:
    """Marks the query as completed."""
    assert self.end_time is None, 'Query is already completed.'

    if error is None:
      # Autofix could lead to a successful `lf.query`, however, the initial
      # lm_response may not be valid. When Error is None, we always try to parse
      # the lm_response into the output. If the output is not valid, the error
      # will be updated accordingly. This logic could be optimized in future by
      # returning attempt information when autofix is enabled.
      if self.schema is not None:
        try:
          output = query_output(
              lm_response, self.schema,
              default=self.default, protocol=self.protocol
          )
        except mapping.MappingError as e:
          output = None
          error = pg.ErrorInfo.from_exception(e)
        self._output = output
      else:
        assert lm_response is not None
        self._output = lm_response.text
    elif (error.tag.startswith('MappingError')
          and self.default != lf.RAISE_IF_HAS_ERROR):
      self._output = self.default

    self.rebind(
        lm_response=lm_response,
        error=error,
        end_time=time.time(),
        skip_notification=True,
    )
    if usage_summary is not None:
      self.usage_summary.merge(usage_summary)

    # Refresh the tab control.
    if self._tab_control is None:
      return

    self._tab_control.insert(
        'schema',
        pg.views.html.controls.Tab(   # pylint: disable=g-long-ternary
            'output',
            pg.view(self.output, collapse_level=None),
            name='output',
        ),
    )
    if self.error is not None:
      self._tab_control.insert(
          'schema',
          pg.views.html.controls.Tab(
              'error',
              pg.view(self.error, collapse_level=None),
              name='error',
          )
      )
    if self.lm_response is not None:
      self._tab_control.append(
          pg.views.html.controls.Tab(
              'lm_response',
              pg.view(
                  self.lm_response,
                  extra_flags=dict(include_message_metadata=True)
              ),
              name='lm_response',
          )
      )
    self._tab_control.select(['error', 'output', 'lm_response'])

  def _html_tree_view_summary(
      self,
      *,
      view: pg.views.HtmlTreeView,
      **kwargs: Any
  ) -> pg.Html | None:
    kwargs.pop('title', None)
    kwargs.pop('enable_summary_tooltip', None)
    return view.summary(
        value=self,
        title=pg.Html.element(
            'div',
            [
                pg.views.html.controls.Label(
                    'lf.query',
                    tooltip=f'[{self.id}] Query invocation',
                    css_classes=['query-invocation-type-name']
                ),
                pg.views.html.controls.Badge(
                    f'lm={self.lm.model_id}',
                    pg.format(
                        self.lm,
                        verbose=False,
                        python_format=True,
                        hide_default_values=True
                    ),
                    css_classes=['query-invocation-lm']
                ),
                pg.views.html.controls.Badge(
                    f'{int(self.elapse)} seconds',
                    css_classes=['query-invocation-time']
                ),
                self.usage_summary.to_html(
                    extra_flags=dict(as_badge=True)
                ),
            ],
            css_classes=['query-invocation-title']
        ),
        enable_summary_tooltip=False,
        **kwargs
    )

  def _html_tree_view_content(
      self,
      *,
      view: pg.views.HtmlTreeView,
      extra_flags: dict[str, Any] | None = None,
      **kwargs: Any
  ) -> pg.Html:
    extra_flags = extra_flags or {}
    interactive = extra_flags.get('interactive', True)
    tab_control = pg.views.html.controls.TabControl([
        pg.views.html.controls.Tab(
            'input',
            pg.view(self.input, collapse_level=None),
            name='input',
        ),
        pg.views.html.controls.Tab(   # pylint: disable=g-long-ternary
            'output',
            pg.view(self.output, collapse_level=None),
            name='output',
        ) if self.is_completed else None,
        pg.views.html.controls.Tab(   # pylint: disable=g-long-ternary
            'error',
            pg.view(self.error, collapse_level=None),
            name='error',
        ) if self.has_error else None,
        pg.views.html.controls.Tab(
            'schema',
            pg.view(self.schema),
            name='schema',
        ),
        pg.views.html.controls.Tab(
            'lm_request',
            pg.view(
                self.lm_request,
                extra_flags=dict(include_message_metadata=False),
            ),
            name='lm_request',
        ),
        pg.views.html.controls.Tab(  # pylint: disable=g-long-ternary
            'lm_response',
            pg.view(
                self.lm_response,
                extra_flags=dict(include_message_metadata=True)
            ),
            name='lm_response',
        ) if self.is_completed else None,
    ], tab_position='top', selected=1)
    if interactive:
      self._tab_control = tab_control
    return tab_control.to_html(extra_flags=extra_flags)

  @classmethod
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .query-invocation-title {
          display: inline-block;
          font-weight: normal;
        }
        .query-invocation-type-name {
          color: #888;
        }
        .query-invocation-lm.badge {
          margin-left: 5px;
          margin-right: 5px;
          color: white;
          background-color: mediumslateblue;
        }
        .query-invocation-time.badge {
          margin-left: 5px;
          border-radius: 0px;
          font-weight: bold;
          background-color: aliceblue;
        }
        .query-invocation-title .usage-summary.label {
          border-radius: 0px;
          color: #AAA;
        }
        """
    ]


@dataclasses.dataclass
class _QueryTracker:
  """Query tracker for `track_queries`."""

  include_child_scopes: Annotated[
      bool,
      (
          'If True, the queries made in nested `track_queries` contexts will '
          'be tracked by this tracker. Otherwise, only the queries made in the '
          'current scope will be included.'
      )
  ] = True

  start_callabck: Annotated[
      Callable[[QueryInvocation], None] | None,
      (
          'A callback function to be called when a query is started.'
      )
  ] = None

  end_callabck: Annotated[
      Callable[[QueryInvocation], None] | None,
      (
          'A callback function to be called when a query is completed.'
      )
  ] = None

  tracked_queries: Annotated[
      list[QueryInvocation],
      (
          'The list of queries tracked by this tracker.'
      )
  ] = dataclasses.field(default_factory=list)

  def track(self, invocation: QueryInvocation) -> None:
    self.tracked_queries.append(invocation)
    if self.start_callabck is not None:
      self.start_callabck(invocation)

  def mark_completed(self, invocation: QueryInvocation) -> None:
    assert invocation in self.tracked_queries, invocation
    if self.end_callabck is not None:
      self.end_callabck(invocation)


@contextlib.contextmanager
def track_queries(
    include_child_scopes: bool = True,
    *,
    start_callabck: Callable[[QueryInvocation], None] | None = None,
    end_callabck: Callable[[QueryInvocation], None] | None = None,
) -> Iterator[list[QueryInvocation]]:
  """Track all queries made during the context.

  Example:

    ```
    with lf.track_queries() as queries:
      lf.query('hi', lm=lm)
      lf.query('What is this {{image}}?', lm=lm, image=image)

    print(queries)
    ```

  Args:
    include_child_scopes: If True, the queries made in child scopes will be
      included in the returned list. Otherwise, only the queries made in the
      current scope will be included.
    start_callabck: A callback function to be called when a query is started.
    end_callabck: A callback function to be called when a query is completed.

  Yields:
    A list of `QueryInvocation` objects representing the queries made during
    the context.
  """
  trackers = lf.context_value('__query_trackers__', [])
  tracker = _QueryTracker(
      include_child_scopes=include_child_scopes,
      start_callabck=start_callabck,
      end_callabck=end_callabck
  )

  with lf.context(
      __query_trackers__=[tracker] + trackers
  ):
    try:
      yield tracker.tracked_queries
    finally:
      pass



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""LangFunc: Language-based functions."""

import dataclasses
from typing import Annotated, Type

from langfun.core import component
from langfun.core import language_model
from langfun.core import message as message_lib
from langfun.core import subscription
from langfun.core import template as template_lib
import pyglove as pg


# Keys for storing render/call stack (invoked LangFunc objects) in the
# thread-local storage.
_TLS_LFUN_CALL_STACK = '_langfunc_callstack'


# NOTE(daiyip): Only the template string belongs to the positional arguments,
# all others are keyword-only for clarity.
@pg.use_init_args(['template_str'])
class LangFunc(
    template_lib.Template,
):
  r"""Base class for natural-language driven component.

  ``LangFunc`` is a language-driven component that enables users to
  seamlessly interact with Language Models (LLMs) using a blend of natural
  language and code. It empowers users to easily modularize prompt/execution
  logics, compose them, and simplify the creation of Language Model (LLM)-based
  components and applications.

  LangFunc can be conceptualized as a string template with embeddable code,
  but it distinguishes itself from traditional template systems in four key
  ways.

  Firstly, it enables easy modularization of templates along with the required
  values with OO principles, providing a reusable way for LLM-based content
  generation. For example:

    ```
    class FewshotExamples(lf.LangFunc):
      '''Base for fewshot prompt.

      {% for example in examples %}
      {{ example }}
      {% endfor %}
      '''

    # Usage 1: __init__ time binding.
    assert FewshotPrompt(examples=['foo', 'bar'])() == 'foo\nbar'

    # Usage 2: __call__ time binding.
    assert FewshotPrompt()(examples=['foo', 'bar']) == 'foo\nbar'

    class ToolDescription(lf.LangFunc):
      '''Tool descriptions.

      {% for tool in tools %}
      {{ tool.description }}
      {% endfor %}
      '''
      # We want to constrain tools to be a list of `Tool` objects.
      tools: list[Tool]

    # Raises: runtime type checking will fail on [1, 2, 3].
    ToolDescription(tools=[1, 2, 3])
    ```

  Secondly, it has the capability to compose multiple LangFuncs together,
  enabling the accomplishment of complex language tasks with maximum reuse.
  It allows users to provide program inputs to all the LangFuncs within a
  composition at the top level, significantly simplifying the process of
  providing context for users. For example:

    ```
    class ReAct(lf.LangFunc):
      '''ReAct prompt for tool-use.

      {{ preamble }}
      {{ tool_description }}
      {{ tool_examples }}
      {{ user_input }}
      '''
      # Default preamble, which could be overriden from subclass
      # or parsed from the `__init__` argument.
      preamble = 'Please help me on my task based on the following tools.',

    react = ReAct(
        tool_description=ToolDescription()
        tool_examples=FewshotExamples(),
        # Partially bind `tools` and `examples`.
        tools=my_tools,
        examples=[t.examples for t in my_tools]
        )

    # Late bind `user_input` at __call__ time.
    react(user_input='Help me get a lunch to go, veggie please.' )
    ```

  Thirdly, it allows the flexibility to encapsulate complex compositions to
  reusable classes and modify them. For example:

    ```
    # The compound decorator converts a function into a LangFunc.
    @lf.compound
    def react_with_tools(preamble, tools: list[Tool]):
      return ReAct(
          preamble=preamble,
          tool_description=ToolDescription()
          tool_examples=FewshotExamples(),
          # Partially bind `tools` and `examples`.
          tools=my_tools,
          examples=[t.examples for t in my_tools]
      )

    # Actually, the entire chat application is a LangFunc.
    class Chat(lt.LangFunc):
      '''LLM-based Chat application.

      llm({{ prompt }})
      '''

    chat = Chat(
        llm=Bard24B(),
        prompt=react_with_tools(
            preamble=(
                f'Please help me solve my problem using tools. '
                f'Current time is {{datetime.datetime.now()}}'),
            tools=my_tools))

    chat(user_input='Help me get a lunch to go, veggie please.')
    ```

  Fourthly, LangFunc is built on top of PyGlove symbolic programming power,
  it could be manipulated programmatically, turned into a space for data
  sampling, or even tuned by AutoML. For example:

    ```
    import pyglove as pg

    prompt_space = react_with_tools(
        preamble=pg.oneof([
            'Help me solve my problem using the following tools:',
            'Help me with the tools below:',
            ...
        ])
        # Choose any two of the tools for generating data.
        tools=pg.manyof(2, [
            google_search(...),
            doordash(...),
            ...
        ])

    for prompt in pg.random_sample(prompt_space):
      print(prompt(user_input='Help me book a conf room please.'))

    ```

  For more capabilities on symbolic programming with PyGlove, please checkout
  https://pyglove.readthedocs.io/en/latest/.

  Final note: always include these capitalized words if you don't want to treat
  the docstr as the template str: THIS IS NOT A TEMPLATE. So as a result, this
  docstr is not treated as a template str :).
  """

  lm: Annotated[
      language_model.LanguageModel,
      (
          'The language model used to sample the responses based on the '
          'rendered template. It could either be provided at `__init__` time '
          'or accessed at `__call__` time from its context.'
      ),
  ] = component.contextual()

  def _on_bound(self):
    super()._on_bound()

    # Last LM input and output.
    self._cached_lm_input = None
    self._cached_lm_output = None

  @property
  def lm_input(self) -> message_lib.Message | None:
    """Returns the cached LM input from the last invocation to `__call__`."""
    return self._cached_lm_input

  @property
  def lm_output(self) -> message_lib.Message | None:
    """Returns the cached LM output from the last invocation to `__call__`."""
    return self._cached_lm_output

  def __call__(
      self,
      *,
      lm: language_model.LanguageModel | None = None,
      lm_input: message_lib.Message | None = None,
      cache_seed: int | None = 0,
      skip_lm: bool = False,
      **variables,
  ) -> message_lib.Message:
    """Calls language model with `lm_input` or rendered text.

    Args:
      lm: Language model to use. When present, it takes precedence over the
        value from the `lm` attribute as well as from the containing chain.
      lm_input: An optional `lf.Message` object as the `lm_input`. If present,
        it will be directly used to call the LM. Otherwise, `render()` will be
        called to produce an LM input first.
      cache_seed: Seed for computing cache key. The cache key is determined by a
        tuple of (lm, prompt, cache seed). If None, cache will be disabled for
        the query even cache is configured by the LM.
      skip_lm: If True, returns the rendered prompt as a UserMessage object.
        otherwise return the LLM response based on the rendered prompt.
      **variables: Template variables applicable to this or child LangFunc.

    Returns:
      An output message from LM.
    """
    return self._call_once(
        lm=lm,
        lm_input=lm_input,
        cache_seed=cache_seed,
        skip_lm=skip_lm,
        **variables,
    )

  def _call_once(
      self,
      *,
      lm: language_model.LanguageModel | None = None,
      lm_input: message_lib.Message | None = None,
      cache_seed: int | None = 0,
      skip_lm: bool = False,
      **variables,
  ) -> message_lib.Message:
    """Call the language model once, with invoking the output transform."""
    try:
      pg.object_utils.thread_local_push(_TLS_LFUN_CALL_STACK, self)

      kwargs = dict(variables)
      if lm is not None:
        kwargs['lm'] = lm

      with self.override(**kwargs):
        # Render the LM input text and creates a user message.
        if lm_input is None:
          lm_input = self.render(**kwargs)

        if skip_lm:
          return lm_input

        self._cached_lm_input = lm_input

        # Send rendered text to LM.
        lm_output = self.lm(lm_input, cache_seed=cache_seed)

        # Attach cache seed.
        lm_input.metadata.cache_seed = cache_seed

        # Transform the output message.
        lm_output = self.transform_output(lm_output)
        lm_output.tag(message_lib.Message.TAG_LM_OUTPUT)

        # We cache the transformed output instead of the original one
        # since the old one is tracked with `sym_origin`.
        self._cached_lm_output = lm_output

      # Emit LangFuncCallEvent.
      lm_callstack = list(
          pg.object_utils.thread_local_get(_TLS_LFUN_CALL_STACK))
      lm_callstack.pop()
      subscription.emit(
          LangFuncCallEvent(
              sender=self,
              lm_input=lm_input,
              lm_output=lm_output,
              lm_callstack=lm_callstack,
          )
      )
      return lm_output
    finally:
      top = pg.object_utils.thread_local_pop(_TLS_LFUN_CALL_STACK, self)
      assert top is self, (top, self)

  def render(
      self,
      *,
      allow_partial: bool = False,
      implicit: bool = False,
      message_cls: Type[message_lib.Message] = message_lib.UserMessage,
      **kwargs,
  ) -> message_lib.Message:
    lm_input = super().render(
        allow_partial=allow_partial,
        implicit=implicit,
        message_cls=message_cls,
        **kwargs,
    )

    with component.context(**kwargs):
      with self.override(**kwargs):
        return self.transform_input(lm_input)

  #
  # Input and output transforms.
  # Subclasses can override.
  #

  def transform_input(
      self, lm_input: message_lib.Message) -> message_lib.Message:
    """Transforms the input message before sending to LM."""
    return lm_input

  def transform_output(
      self, lm_output: message_lib.Message) -> message_lib.Message:
    """Transforms the output message before returning from __call__."""
    return lm_output


# Register converter from str to LangFunc, therefore we can always
# pass strs to attributes that accept LangFunc.
pg.typing.register_converter(str, LangFunc, LangFunc)


#
# LangFunc events.
#


@dataclasses.dataclass
class LangFuncCallEvent(subscription.Event[LangFunc]):
  """LangFunc call event."""

  lm_input: message_lib.Message
  lm_output: message_lib.Message
  lm_callstack: list[LangFunc]



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Video tests."""
import unittest
from unittest import mock

from langfun.core.modalities import video as video_lib
import pyglove as pg


mp4_bytes = (
    b'\x00\x00\x00\x20ftypmp42'
    b'\x00\x00\x00\x08moov'
    b'\x00\x00\x00\x0ctrak'
    b'\x00\x00\x00\x08mdia'
)


def mock_request(*args, **kwargs):
  del args, kwargs
  return pg.Dict(content=mp4_bytes)


class VideoContentTest(unittest.TestCase):

  def test_video_content(self):
    video = video_lib.Video.from_bytes(mp4_bytes)
    self.assertEqual(video.mime_type, 'video/mp4')
    self.assertEqual(video.video_format, 'mp4')
    self.assertIn('data:video/mp4;base64,', video._raw_html())
    self.assertEqual(video.to_bytes(), mp4_bytes)

  def test_bad_video(self):
    video = video_lib.Video.from_bytes(b'bad')
    with self.assertRaisesRegex(ValueError, 'Expected MIME type'):
      _ = video.video_format


class VideoFileTest(unittest.TestCase):

  def test_video_file(self):
    video = video_lib.Video.from_uri('http://mock/web/a.mp4')
    with mock.patch('requests.get') as mock_requests_get:
      mock_requests_get.side_effect = mock_request
      self.assertEqual(video.video_format, 'mp4')
      self.assertEqual(video.mime_type, 'video/mp4')
      self.assertEqual(
          video._raw_html(),
          '<video controls> <source src="http://mock/web/a.mp4"> </video>',
      )
      self.assertEqual(video.to_bytes(), mp4_bytes)


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility for async IO in Langfun."""

import asyncio
from typing import Any, Callable
import pyglove as pg


async def invoke_async(
    callable_object: Callable[..., Any], *args, **kwargs
) -> Any:
  """Invokes a callable asynchronously with `lf.context` manager enabled."""
  return await asyncio.to_thread(
      # Enable `lf.context` manager for async calls.
      pg.with_contextual_override(callable_object), *args, **kwargs
  )



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Langfun."""

# pylint: disable=g-bad-import-order
# pylint: disable=g-import-not-at-top
# pylint: disable=unused-import
from langfun.core import *
from langfun.core import structured

Schema = structured.Schema
MISSING = structured.MISSING
UNKNOWN = structured.UNKNOWN

include_method_in_prompt = structured.include_method_in_prompt

MappingExample = structured.MappingExample

call = structured.call
parse = structured.parse
query = structured.query
describe = structured.describe
complete = structured.complete
score = structured.score
generate_class = structured.generate_class

track_queries = structured.track_queries

# Context manager for setting the query protocol for the scope.
query_protocol = structured.query_protocol

# Helper function for map-reduce style querying.
query_and_reduce = structured.query_and_reduce

# Helper functions for input/output transformations based on
# `lf.query` (e.g. jax-on-beam could use these for batch processing)
query_prompt = structured.query_prompt
query_output = structured.query_output

source_form = structured.source_form
function_gen = structured.function_gen

from langfun.core import data
from langfun.core import eval  # pylint: disable=redefined-builtin
from langfun.core import templates
from langfun.core import coding

PythonCode = coding.PythonCode
PythonFunction = coding.PythonFunction

from langfun.core import llms
lm_cache = llms.cache.lm_cache

from langfun.core import agentic
Action = agentic.Action
Session = agentic.Session

from langfun.core import memories

from langfun.core import modalities

Mime = modalities.Mime
MimeType = Mime  # For backwards compatibility.
Image = modalities.Image
Video = modalities.Video
PDF = modalities.PDF

# Additional error types.
MappingError = structured.MappingError
SchemaError = structured.SchemaError
JsonError = structured.JsonError
CodeError = coding.CodeError

# Placeholder for Google-internal imports.

# pylint: enable=unused-import
# pylint: enable=g-import-not-at-top
# pylint: enable=g-bad-import-order

__version__ = "0.1.2"



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OpenAI API message conversion."""

from typing import Annotated, Any, Callable

import langfun.core as lf
from langfun.core import modalities as lf_modalities


class OpenAIMessageConverter(lf.MessageConverter):
  """Converter to OpenAI API."""

  FORMAT_ID = 'openai'

  chunk_preprocessor: Annotated[
      Callable[[str | lf.Modality], Any] | None,
      (
          'Chunk preprocessor for Langfun chunk to OpenAI chunk conversion. '
          'It will be applied before each Langfun chunk is converted. '
          'If returns None, the chunk will be skipped.'
      )
  ] = None

  def to_value(self, message: lf.Message) -> dict[str, Any]:
    """Converts a Langfun message to OpenAI API."""
    parts = []
    for chunk in message.chunk():
      if self.chunk_preprocessor is not None:
        chunk = self.chunk_preprocessor(chunk)
        if chunk is None:
          continue

      if isinstance(chunk, str):
        item = dict(type='text', text=chunk)
      elif isinstance(chunk, lf_modalities.Image):
        item = dict(
            type='image_url', image_url=dict(url=chunk.embeddable_uri)
        )
      # TODO(daiyip): Support audio_input.
      else:
        raise ValueError(f'Unsupported content type: {chunk!r}.')
      parts.append(item)
    return dict(
        role=self.get_role(message),
        content=parts,
    )

  def get_role(self, message: lf.Message) -> str:
    """Returns the role of the message."""
    if isinstance(message, lf.SystemMessage):
      return 'system'
    elif isinstance(message, lf.UserMessage):
      return 'user'
    elif isinstance(message, lf.AIMessage):
      return 'assistant'
    else:
      raise ValueError(f'Unsupported message type: {message!r}.')

  def get_message_cls(self, role: str) -> type[lf.Message]:
    """Returns the message class of the message."""
    match role:
      case 'system':
        return lf.SystemMessage
      case 'user':
        return lf.UserMessage
      case 'assistant':
        return lf.AIMessage
      case _:
        raise ValueError(f'Unsupported role: {role!r}.')

  def from_value(self, value: dict[str, Any]) -> lf.Message:
    """Returns a Langfun message from OpenAI message."""
    message_cls = self.get_message_cls(
        self._safe_read(value, 'role', default='assistant')
    )
    content = self._safe_read(value, 'content')
    if isinstance(content, str):
      return message_cls(content)

    assert isinstance(content, list)
    chunks = []
    for item in content:
      t = self._safe_read(item, 'type')
      if t == 'text':
        chunk = self._safe_read(item, 'text')
      elif t == 'image_url':
        chunk = lf_modalities.Image.from_uri(
            self._safe_read(self._safe_read(item, 'image_url'), 'url')
        )
      else:
        raise ValueError(f'Unsupported content type: {item!r}.')
      chunks.append(chunk)
    return message_cls.from_chunks(chunks)


def _as_openai_format(
    self,
    chunk_preprocessor: Callable[[str | lf.Modality], Any] | None = None,
    **kwargs
) -> dict[str, Any]:
  """Returns an OpenAI format message."""
  return OpenAIMessageConverter(
      chunk_preprocessor=chunk_preprocessor, **kwargs
  ).to_value(self)


@classmethod
def _from_openai_format(
    cls,
    openai_message: dict[str, Any],
    **kwargs
) -> lf.Message:
  """Creates a Langfun message from the OpenAI format message."""
  del cls
  return OpenAIMessageConverter(**kwargs).from_value(openai_message)

# Set shortcut methods in lf.Message.
lf.Message.as_openai_format = _as_openai_format
lf.Message.from_openai_format = _from_openai_format



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fake LMs for testing."""

import abc
import functools
from typing import Annotated
import langfun.core as lf


class Fake(lf.LanguageModel):
  """The base class for all fake language models."""

  def _score(self, prompt: lf.Message| list[lf.Message],
             completions: list[lf.Message]):
    return [lf.LMScoringResult(score=-i * 1.0) for i in range(len(completions))]

  def _tokenize(self, prompt: lf.Message) -> list[tuple[str | bytes, int]]:
    return [(w, i) for i, w in enumerate(prompt.text.split(' '))]

  def _sample(self, prompts: list[lf.Message]) -> list[lf.LMSamplingResult]:
    results = []
    for prompt in prompts:
      response = self._response_from(prompt)
      results.append(
          lf.LMSamplingResult(
              [lf.LMSample(response, 1.0)],
              usage=lf.LMSamplingUsage(
                  prompt_tokens=len(prompt.text),
                  completion_tokens=len(response.text),
                  total_tokens=len(prompt.text) + len(response.text),
              )
          )
      )
    return results

  @functools.cached_property
  def model_info(self) -> lf.ModelInfo:
    """Returns the specification of the model."""
    return lf.ModelInfo(model_id=self.__class__.__name__)

  @abc.abstractmethod
  def _response_from(self, prompt: lf.Message) -> lf.Message:
    """Returns the response for the given prompt."""


class Echo(Fake):
  """A simple echo language model for testing."""

  def _response_from(self, prompt: lf.Message) -> lf.Message:
    return lf.AIMessage(prompt.text)


class Pseudo(Fake):
  """A pseudo language model that should never be called."""

  def _response_from(self, prompt: lf.Message) -> lf.Message:
    raise ValueError('Pseudo language model should never be called.')


@lf.use_init_args(['response'])
class StaticResponse(Fake):
  """Language model that always gives the same canned response."""

  response: Annotated[
      str | lf.Message,
      'A canned response that will be returned regardless of the prompt.'
  ]

  def _response_from(self, prompt: lf.Message) -> lf.Message:
    return lf.AIMessage.from_value(self.response)


@lf.use_init_args(['mapping'])
class StaticMapping(Fake):
  """A static mapping from prompt to response."""

  mapping: Annotated[
      dict[str, str | lf.Message],
      'A mapping from prompt to response.'
  ]

  def _response_from(self, prompt: lf.Message) -> lf.Message:
    return lf.AIMessage.from_value(self.mapping[prompt])


@lf.use_init_args(['sequence'])
class StaticSequence(Fake):
  """A static sequence of responses to use."""

  sequence: Annotated[
      list[str | lf.Message],
      'A sequence of strings as the response.'
  ]

  def _on_bound(self):
    super()._on_bound()
    self._pos = 0

  def _response_from(self, prompt: lf.Message) -> lf.Message:
    r = lf.AIMessage.from_value(self.sequence[self._pos])
    self._pos += 1
    return r



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base classes for Langfun evaluation."""

import dataclasses
import inspect
from typing import Any, Callable, Iterator
import langfun.core as lf
import pyglove as pg


@dataclasses.dataclass
class Example(pg.JSONConvertible, pg.views.HtmlTreeView.Extension):
  """An item for the evaluation.

  Attributes:
    id: The 1-based ID of the item in the evaluation set.
    input: An element returned from the `Evaluable.inputs` functor.
    output: The output of the `process` method. If `pg.MISSING_VALUE`, it has
      not been processed yet.
    metadata: The metadata of the item produced by the `process` method.
    metric_metadata: The dictionary returned from `Metric.audit`.
    start_time: The start time of the evaluation item.
    end_time: The end time of the evaluation item.
    usage_summary: The summary of LLM usages of the evaluation item.
    execution_status: The timeit status of the evaluation item.
  """
  id: int
  input: Any = pg.MISSING_VALUE
  output: Any = pg.MISSING_VALUE
  error: pg.ErrorInfo | None = None
  metadata: dict[str, Any] = dataclasses.field(default_factory=dict)
  metric_metadata: dict[str, Any] | None = None
  # Execution information.
  newly_processed: bool = True
  start_time: float | None = None
  end_time: float | None = None
  usage_summary: lf.UsageSummary | None = None
  execution_status: dict[str, pg.utils.TimeIt.Status] | None = None

  def __post_init__(self):
    if self.execution_status is not None:
      for status in self.execution_status.values():
        if status.has_error:
          assert isinstance(status.error, pg.ErrorInfo)
          self.error = status.error
          break

  @property
  def is_processed(self) -> bool:
    """Returns whether the item has been processed."""
    return pg.MISSING_VALUE != self.output

  @property
  def has_error(self) -> bool:
    """Returns whether the item has an error."""
    return self.error is not None

  @property
  def elapse(self) -> float | None:
    """Returns the elapse time of the item."""
    if self.execution_status is not None:
      return self.execution_status['evaluate'].elapse
    return None

  def to_json(self, *, exclude_input: bool = False, **kwargs):
    """Returns the JSON representation of the item."""
    return self.to_json_dict(
        fields=dict(
            id=(self.id, None),
            input=(
                self.input if not exclude_input else pg.MISSING_VALUE,
                pg.MISSING_VALUE
            ),
            output=(self.output, pg.MISSING_VALUE),
            error=(self.error, None),
            metadata=(self.metadata, {}),
            metric_metadata=(self.metric_metadata, None),
            start_time=(self.start_time, None),
            end_time=(self.end_time, None),
            usage_summary=(self.usage_summary, None),
            execution_status=(self.execution_status, None),
        ),
        exclude_default=True,
        **kwargs,
    )

  @classmethod
  def from_json(
      cls,
      json_value: dict[str, Any],
      *,
      example_input_by_id: Callable[[int], Any] | None = None,
      load_example_metadata: bool | Callable[['Example'], bool] = False,
      **kwargs
  ) -> 'Example':
    """Creates an example from the JSON representation."""
    example_id = json_value.get('id')
    if example_input_by_id:
      example_input = example_input_by_id(example_id)
    else:
      example_input = json_value.pop('input', pg.MISSING_VALUE)
      if example_input is not pg.MISSING_VALUE:
        example_input = pg.from_json(example_input, **kwargs)
    json_value['input'] = example_input

    # NOTE(daiyip): We need to load the types of the examples into the
    # deserialization context, otherwise the deserialization will fail if the
    # types are not registered.
    def example_class_defs(example) -> list[type[Any]]:
      referred_types = set()
      def _visit(k, v, p):
        del k, p
        if inspect.isclass(v):
          referred_types.add(v)
        elif isinstance(v, pg.Object):
          referred_types.add(v.__class__)
        return pg.TraverseAction.ENTER
      pg.traverse(example, _visit)
      return list(referred_types)

    # We delay loading the metadata until the other parts of the example are
    # loaded. So we could apply the filter to decide whether to load the
    # metadata.
    metadata_dict = json_value.pop('metadata', None)
    with pg.JSONConvertible.load_types_for_deserialization(
        *example_class_defs(example_input)
    ):
      example = cls(
          **{k: pg.from_json(v, **kwargs) for k, v in json_value.items()}
      )
      if callable(load_example_metadata):
        load_example_metadata = load_example_metadata(example)
      if load_example_metadata:
        example.metadata = pg.from_json(metadata_dict, **kwargs)
      return example

  @classmethod
  def iter_ckpts(
      cls,
      ckpt_file: str | list[str],
      example_input_by_id: Callable[[int], Any] | None = None,
      load_example_metadata: bool = True,
  ) -> Iterator['Example']:
    """Iterates Examples from the checkpoint files."""
    ckpt_files = [ckpt_file] if isinstance(ckpt_file, str) else ckpt_file
    for ckpt_file in ckpt_files:
      with pg.io.open_sequence(ckpt_file) as f:
        for record in f:
          example = pg.from_json_str(
              record,
              example_input_by_id=example_input_by_id,
              load_example_metadata=load_example_metadata
          )
          assert isinstance(example, cls), example
          yield example

  #
  # HTML rendering.
  #

  def _html_tree_view_content(
      self,
      *,
      view: pg.views.HtmlTreeView,
      root_path: pg.KeyPath | None = None,
      extra_flags: dict[str, Any] | None = None,
      **kwargs
  ):
    root_path = root_path or pg.KeyPath()
    extra_flags = extra_flags or {}
    num_examples = extra_flags.get('num_examples', None)

    def _metric_metadata_badge(key, value):
      if isinstance(value, bool) and bool:
        text = key
      else:
        text = f'{key}:{value}'
      return pg.views.html.controls.Badge(
          text,
          css_classes=[pg.utils.camel_to_snake(key, '-')],
      )

    def _render_header():
      return pg.Html.element(
          'div',
          [
              pg.Html.element(
                  'div',
                  [
                      # Previous button.
                      pg.views.html.controls.Label(   # pylint: disable=g-long-ternary
                          '◀',
                          link=f'{self.id - 1}.html',
                          css_classes=['previous'],
                      ) if self.id > 1 else None,
                      # Current example ID.
                      pg.views.html.controls.Label(
                          f'#{self.id}',
                          css_classes=['example-id'],
                      ),
                      # Next button.
                      pg.views.html.controls.Label(   # pylint: disable=g-long-ternary
                          '▶',
                          link=f'{self.id + 1}.html',
                          css_classes=['next'],
                      ) if (num_examples is None
                            or self.id < num_examples) else None,

                  ]
              ),
              pg.Html.element(
                  'div',
                  [
                      # Usage summary.
                      pg.view(  # pylint: disable=g-long-ternary
                          self.usage_summary,
                          extra_flags=dict(as_badge=True)
                      ) if self.usage_summary is not None else None,
                      # Metric metadata.
                      pg.views.html.controls.LabelGroup(
                          [   # pylint: disable=g-long-ternary
                              _metric_metadata_badge(k, v)
                              for k, v in self.metric_metadata.items()
                          ] if self.metric_metadata else []
                      ),
                  ],
                  css_classes=['example-container'],
              )
          ]
      )

    def _render_content():
      def _tab(label, key, default):
        field = getattr(self, key)
        if default == field:
          return None
        return pg.views.html.controls.Tab(
            label=label,
            content=view.render(
                field,
                root_path=root_path + key,
                collapse_level=None,
                **view.get_passthrough_kwargs(**kwargs),
            ),
        )
      tabs = [
          _tab('Input', 'input', pg.MISSING_VALUE),
          _tab('Output', 'output', pg.MISSING_VALUE),
          _tab('Output Metadata', 'metadata', {}),
          _tab('Error', 'error', None),
      ]
      tabs = [tab for tab in tabs if tab is not None]
      return pg.views.html.controls.TabControl(
          tabs,
          len(tabs) - 1,
      )

    return pg.Html.element(
        'div',
        [
            _render_header(),
            _render_content(),
        ],
        css_classes=['eval-example']
    )

  def _html_tree_view_summary(self, *, view, **kwargs):
    return None

  @classmethod
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .example-container {
          display: block;
          padding: 10px;
        }
        .example-id {
          font-weight: bold;
          font-size: 40px;
          margin: 0 10px;
          vertical-align: middle;
        }
        a.previous, a.next {
          text-decoration: none;
          vertical-align: middle;
          display: inline-block;
          padding: 8px 8px;
          color: #DDD;
        }
        a.previous:hover, a.next:hover {
          background-color: #ddd;
          color: black;
        }
        /* Badge styles. */
        .eval-example .badge.match {
          color: green;
          background-color: #dcefbe;
        }
        .eval-example .badge.error {
          color: red;
          background-color: #fdcccc;
        }
        .eval-example .badge.mismatch {
          color: orange;
          background-color: #ffefc4;
        }
        .eval-example .badge.score {
          color: blue;
          background-color: #c4dced;
        }
        """
    ]




# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib
import io
import os
import tempfile
import unittest

from langfun.core import console as lf_console
from langfun.core.eval.v2 import eval_test_helper
from langfun.core.eval.v2 import progress_tracking    # pylint: disable=unused-import
from langfun.core.eval.v2 import runners as runners_lib  # pylint: disable=unused-import
import pyglove as pg


class HtmlProgressTrackerTest(unittest.TestCase):

  def test_track_progress(self):
    result = pg.Dict()
    def display(x):
      result['view'] = x.to_html()

    lf_console._notebook = pg.Dict(
        display=display
    )
    root_dir = os.path.join(tempfile.mkdtemp(), 'test_html_progress_tracker')
    experiment = eval_test_helper.test_experiment()
    _ = experiment.run(root_dir, 'new', plugins=[])
    self.assertIsInstance(result['view'], pg.Html)
    lf_console._notebook = None


class TqdmProgressTrackerTest(unittest.TestCase):

  def test_basic(self):
    root_dir = os.path.join(tempfile.mkdtemp(), 'test_tqdm_progress_tracker')
    experiment = eval_test_helper.test_experiment()
    string_io = io.StringIO()
    with contextlib.redirect_stderr(string_io):
      _ = experiment.run(root_dir, 'new', plugins=[])
    self.assertIn('All: 100%', string_io.getvalue())

  def test_with_example_ids(self):
    root_dir = os.path.join(
        tempfile.mkdtemp(), 'test_tqdm_progress_tracker_with_example_ids'
    )
    experiment = eval_test_helper.test_experiment()
    string_io = io.StringIO()
    with contextlib.redirect_stderr(string_io):
      _ = experiment.run(root_dir, 'new', example_ids=[1], plugins=[])
    self.assertIn('All: 100%', string_io.getvalue())


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Google GenAI models."""

import os
import unittest
import langfun.core as lf
from langfun.core.llms import google_genai


class GenAITest(unittest.TestCase):
  """Tests for GenAI model."""

  def test_basics(self):
    with self.assertRaisesRegex(ValueError, 'Please specify `api_key`'):
      _ = google_genai.Gemini15Pro().api_endpoint

    self.assertIsNotNone(google_genai.Gemini15Pro(api_key='abc').api_endpoint)

    os.environ['GOOGLE_API_KEY'] = 'abc'
    lm = google_genai.Gemini15Pro_001()
    self.assertIsNotNone(lm.api_endpoint)
    self.assertEqual(lm.model_id, 'gemini-1.5-pro-001')
    self.assertEqual(lm.resource_id, 'google_genai://gemini-1.5-pro-001')
    del os.environ['GOOGLE_API_KEY']

  def test_lm_get(self):
    self.assertIsInstance(
        lf.LanguageModel.get('google_genai://gemini-1.5-pro'),
        google_genai.GenAI,
    )

if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for structures for Python code generation."""

import inspect
import unittest
from langfun.core.coding.python import generation
import pyglove as pg


class PythonCodeTest(unittest.TestCase):

  def test_auto_run(self):
    with generation.PythonCode.auto_run():
      self.assertEqual(
          generation.PythonCode(
              """
              x = 1
              y = x + 1
              z = x + y
              """,
          ),
          3,
      )
      with generation.PythonCode.auto_run(False):
        self.assertIsInstance(generation.PythonCode('1'), generation.PythonCode)

  def test_eval(self):
    self.assertEqual(
        generation.PythonCode("""
            x = 1
            y = x + 1
            print(y)
            z = x + y
            """).eval(),
        dict(x=1, y=2, z=3, __result__=3, __stdout__='2\n'),
    )

  def test_call(self):
    self.assertEqual(
        generation.PythonCode("""
        x = 1
        y = x + 1
        z = x + y
        """)(),
        3,
    )

  def test_call_class_def(self):
    with pg.coding.permission(pg.coding.CodePermission.CLASS_DEFINITION):
      v = generation.PythonCode("""
          class A:
            pass
          """)(sandbox=False)
      self.assertTrue(inspect.isclass(v))


class PythonFunctionTest(unittest.TestCase):

  def test_basic(self):
    f = generation.PythonFunction(
        name='sum',
        args=dict(x='int', y='int'),
        returns='int',
        source=("""
            def sum(x: int, y: int):
              return x + y
            """),
    )
    self.assertEqual(f(1, y=2), 3)
    self.assertEqual(f(1, y=2, sandbox=False), 3)

  def test_bad_code(self):
    f = generation.PythonFunction(
        name='sum',
        args=dict(x='int', y='int'),
        returns='int',
        source=("""
            def sum(x: int, y: int):
              s = 0
              for _ in range(x):
                s += 1
              for _ in range(y):
                s += 1
              return s
            """),
    )
    with self.assertRaises(TimeoutError):
      f(100000000, y=10000000, timeout=1)


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for message."""

import inspect
import unittest
from langfun.core import language_model
from langfun.core import message
from langfun.core import modality
import pyglove as pg


class CustomModality(modality.Modality):
  content: str

  def to_bytes(self):
    return self.content.encode()


class MessageTest(unittest.TestCase):

  def test_basics(self):

    class A(pg.Object):
      pass

    d = pg.Dict(x=A())

    m = message.UserMessage(
        'hi',
        metadata=dict(x=1), x=pg.Ref(d.x),
        y=2,
        tags=['lm-input']
    )
    self.assertEqual(m.metadata, {'x': pg.Ref(d.x), 'y': 2})
    self.assertEqual(m.sender, 'User')
    self.assertIs(m.x, d.x)
    self.assertEqual(m.y, 2)
    self.assertTrue(m.has_tag('lm-input'))
    self.assertTrue(m.has_tag(('lm-input', '')))
    self.assertFalse(m.has_tag('lm-response'))

    with self.assertRaises(AttributeError):
      _ = m.z
    self.assertEqual(hash(m), hash(m.text))
    del d

  def test_from_value(self):
    self.assertTrue(
        pg.eq(message.UserMessage.from_value('hi'), message.UserMessage('hi'))
    )
    self.assertTrue(
        pg.eq(
            message.UserMessage.from_value(CustomModality('foo')),
            message.UserMessage('<<[[object]]>>', object=CustomModality('foo')),
        )
    )
    m = message.UserMessage('hi')
    self.assertIs(message.UserMessage.from_value(m), m)

  def test_source_tracking(self):
    m1 = message.UserMessage('hi')
    m1.tag('lm-input')
    self.assertIsNone(m1.source)
    self.assertIs(m1.root, m1)

    m2 = message.UserMessage('foo', source=m1)
    m2.source = m1
    self.assertIs(m2.source, m1)
    self.assertIs(m2.root, m1)
    m2.tag('lm-response')

    m3 = message.UserMessage('bar', source=m2)
    self.assertIs(m3.source, m2)
    self.assertIs(m3.root, m1)
    m3.tag('transformed')
    m3.tag('lm-output')

    self.assertEqual(
        m3.trace(), [m1, m2, m3],
    )
    self.assertEqual(
        m3.trace('lm-input'), [m1]
    )
    self.assertEqual(
        m3.trace('transformed'), [m3]
    )
    self.assertIs(m2.lm_input, m1)
    self.assertIs(m3.lm_input, m1)
    self.assertEqual(m3.lm_inputs, [m1])
    self.assertIs(m2.lm_response, m2)
    self.assertIs(m3.lm_response, m2)
    self.assertEqual(m3.lm_responses, [m2])
    self.assertIs(m3.lm_output, m3)
    self.assertEqual(m3.lm_outputs, [m3])
    self.assertIsNone(m3.last('non-exist'))

  def test_result(self):
    m = message.UserMessage('hi', x=1, y=2)
    self.assertIsNone(m.result)
    m.result = 1
    self.assertEqual(m.result, 1)

  def test_jsonify(self):
    m = message.UserMessage('hi', result=1)
    self.assertEqual(pg.from_json_str(m.to_json_str()), m)

  def test_get(self):

    class A(pg.Object):
      p: int

    # Create a symbolic object and assign it to a container, so we could test
    # pg.Ref.
    a = A(1)
    d = pg.Dict(x=a)

    m = message.UserMessage('hi', x=pg.Ref(a), y=dict(z=[0, 1, 2]))
    self.assertEqual(m.get('text'), 'hi')
    self.assertIs(m.get('x'), a)
    self.assertIs(m.get('x.p'), 1)
    self.assertEqual(m.get('y'), dict(z=[0, 1, 2]))
    self.assertEqual(m.get('y.z'), [0, 1, 2])
    self.assertEqual(m.get('y.z[0]'), 0)
    self.assertIsNone(m.get('p'))
    self.assertEqual(m.get('p', default='foo'), 'foo')
    del d

  def test_set(self):
    m = message.UserMessage('hi', metadata=dict(x=1, z=0))
    m.set('text', 'hello')
    m.set('x', 2)
    m.set('y', [0, 1, 2])
    m.set('y[0]', 1)
    m.set('y[2]', pg.MISSING_VALUE)  # delete `y[2]`.
    m.set('z', pg.MISSING_VALUE)  # delete `z`.
    self.assertEqual(
        m, message.UserMessage('hello', metadata=dict(x=2, y=[1, 1]))
    )

  def test_updates(self):
    m = message.UserMessage('hi')
    self.assertFalse(m.modified)
    self.assertFalse(m.has_errors)

    with m.update_scope():
      m.metadata.x = 1
      m.metadata.y = 1
      self.assertTrue(m.modified)
      self.assertEqual(len(m.updates), 2)
      self.assertFalse(m.has_errors)

      with m.update_scope():
        m.metadata.y = 2
        m.metadata.z = 2
        m.errors.append(ValueError('b'))
        self.assertTrue(m.modified)
        self.assertEqual(len(m.updates), 2)
        self.assertTrue(m.has_errors)
        self.assertEqual(len(m.errors), 1)

        with m.update_scope():
          self.assertFalse(m.modified)
          self.assertFalse(m.has_errors)

    self.assertTrue(m.modified)
    self.assertEqual(len(m.updates), 3)
    self.assertTrue(m.has_errors)
    self.assertEqual(len(m.errors), 1)

    m2 = message.UserMessage('hi')
    m2.apply_updates(m.updates)
    self.assertEqual(m, m2)

  def test_user_message(self):
    m = message.UserMessage('hi')
    self.assertEqual(m.text, 'hi')
    self.assertEqual(m.sender, 'User')
    self.assertEqual(str(m), m.text)

    m = message.UserMessage('hi', sender='Tom')
    self.assertEqual(m.sender, 'Tom')
    self.assertEqual(str(m), m.text)

  def test_ai_message(self):
    m = message.AIMessage('hi')
    self.assertEqual(m.text, 'hi')
    self.assertEqual(m.sender, 'AI')
    self.assertEqual(str(m), m.text)

    m = message.AIMessage('hi', sender='Model')
    self.assertEqual(m.sender, 'Model')
    self.assertEqual(str(m), m.text)

  def test_system_message(self):
    m = message.SystemMessage('hi')
    self.assertEqual(m.text, 'hi')
    self.assertEqual(m.sender, 'System')
    self.assertEqual(str(m), m.text)

    m = message.SystemMessage('hi', sender='Environment1')
    self.assertEqual(m.sender, 'Environment1')
    self.assertEqual(str(m), m.text)

  def test_memory_record(self):
    m = message.MemoryRecord('hi')
    self.assertEqual(m.text, 'hi')
    self.assertEqual(m.sender, 'Memory')
    self.assertEqual(str(m), m.text)

    m = message.MemoryRecord('hi', sender="Someone's Memory")
    self.assertEqual(m.sender, 'Someone\'s Memory')
    self.assertEqual(str(m), m.text)

  def test_get_modality(self):
    m1 = message.UserMessage(
        'hi, this is a {{img1}} and {{x.img2}}',
        img1=CustomModality('foo'),
        x=dict(img2=pg.Ref(CustomModality('bar'))),
    )
    self.assertIs(m1.get_modality('img1'), m1.img1)
    self.assertIs(m1.get_modality('x.img2'), m1.x.img2)
    self.assertIsNone(m1.get_modality('video'))

    m2 = message.SystemMessage('class Question:\n  image={{img1}}', source=m1)
    self.assertIs(m2.get_modality('img1'), m1.img1)
    # We could get the modality object even it's not directly used by current
    # message.
    self.assertIs(m2.get_modality('x.img2'), m1.x.img2)
    self.assertIsNone(m2.get_modality('video'))

    m3 = message.AIMessage(
        'This is the {{output_image}} based on {{x.img2}}',
        output_image=CustomModality('bar'),
        source=m2,
    )
    self.assertIs(m3.get_modality('x.img2'), m1.x.img2)
    self.assertIs(m3.get_modality('output_image'), m3.output_image)
    self.assertIsNone(m3.get_modality('video'))

  def test_referred_modalities(self):
    m1 = message.UserMessage(
        'hi, this is a <<[[img1]]>> and <<[[x.img2]]>>',
        img1=CustomModality('foo'),
        x=dict(img2=CustomModality('bar')),
    )
    m2 = message.SystemMessage('class Question:\n  image={{img1}}', source=m1)
    m3 = message.AIMessage(
        (
            'This is the <<[[output_image]]>> based on <<[[x.img2]]>>, '
            '{{unknown_var}}'
        ),
        output_image=CustomModality('bar'),
        source=m2,
    )
    self.assertEqual(
        m3.referred_modalities(),
        {
            'output_image': m3.output_image,
            'x.img2': m1.x.img2,
        },
    )

  def test_text_with_modality_hash(self):
    m = message.UserMessage(
        'hi, this is a <<[[img1]]>> and <<[[x.img2]]>>',
        img1=CustomModality('foo'),
        x=dict(img2=CustomModality('bar')),
    )
    self.assertEqual(
        m.text_with_modality_hash,
        (
            'hi, this is a <<[[img1]]>> and <<[[x.img2]]>>'
            '<img1>acbd18db</img1><x.img2>37b51d19</x.img2>'
        )
    )

  def test_chunking(self):
    m = message.UserMessage(
        inspect.cleandoc("""
            Hi, this is <<[[a]]>> and this is {{b}}.
            <<[[x.c]]>> {{something else
            """),
        a=CustomModality('foo'),
        x=dict(c=CustomModality('bar')),
    )
    chunks = m.chunk()
    self.assertTrue(
        pg.eq(
            chunks,
            [
                'Hi, this is',
                CustomModality('foo'),
                'and this is {{b}}.\n',
                CustomModality('bar'),
                '{{something else',
            ],
        )
    )
    self.assertTrue(
        pg.eq(
            message.AIMessage.from_chunks(chunks),
            message.AIMessage(
                inspect.cleandoc("""
                    Hi, this is <<[[obj0]]>> and this is {{b}}.
                    <<[[obj1]]>> {{something else
                    """),
                obj0=pg.Ref(m.a),
                obj1=pg.Ref(m.x.c),
            ),
        )
    )

  def assert_html_content(self, html, expected):
    expected = inspect.cleandoc(expected).strip()
    actual = html.content.strip()
    if actual != expected:
      print(actual)
    self.assertEqual(actual, expected)

  def test_html_style(self):
    self.assertIn(
        inspect.cleandoc(
            """
            /* Langfun Message styles.*/
            [class^="message-"] > details {
                margin: 0px 0px 5px 0px;
                border: 1px solid #EEE;
            }
            .lf-message.summary-title::after {
                content: ' 💬';
            }
            details.pyglove.ai-message {
                border: 1px solid blue;
                color: blue;
            }
            details.pyglove.user-message {
                border: 1px solid green;
                color: green;
            }
            .message-tags {
                margin: 5px 0px 5px 0px;
                font-size: .8em;
            }
            .message-tags > span {
                border-radius: 5px;
                background-color: #CCC;
                padding: 3px;
                margin: 0px 2px 0px 2px;
                color: white;
            }
            .message-text {
                padding: 20px;
                margin: 10px 5px 10px 5px;
                font-style: italic;
                white-space: pre-wrap;
                border: 1px solid #EEE;
                border-radius: 5px;
                background-color: #EEE;
            }
            .modality-in-text {
                display: inline-block;
            }
            .modality-in-text > details.pyglove {
                display: inline-block;
                font-size: 0.8em;
                border: 0;
                background-color: #A6F1A6;
                margin: 0px 5px 0px 5px;
            }
            .message-result {
                color: dodgerblue;
            }
            .message-usage {
                color: orange;
            }
            .message-usage .object-key.str {
                border: 1px solid orange;
                background-color: orange;
                color: white;
            }
            """
        ),
        message.UserMessage('hi').to_html().style_section,
    )

  def test_html_user_message(self):
    self.assert_html_content(
        message.UserMessage(
            'what is a <div>'
        ).to_html(enable_summary_tooltip=False),
        """
        <details open class="pyglove user-message lf-message"><summary><div class="summary-title lf-message">UserMessage(...)</div></summary><div class="complex_value"><div class="message-tags"></div><div class="message-text">what is a &lt;div&gt;</div><div class="message-metadata"><details open class="pyglove dict message-metadata"><summary><div class="summary-name message-metadata">metadata<span class="tooltip message-metadata">metadata</span></div><div class="summary-title message-metadata">Dict(...)</div></summary><div class="complex-value dict"><span class="empty-container"></span></div></details></div></div></details>
        """
    )
    self.assert_html_content(
        message.UserMessage(
            'what is this <<[[image]]>>',
            tags=['lm-input'],
            image=CustomModality('bird')
        ).to_html(
            enable_summary_tooltip=False,
            extra_flags=dict(include_message_metadata=False)
        ),
        """
        <details open class="pyglove user-message lf-message"><summary><div class="summary-title lf-message">UserMessage(...)</div></summary><div class="complex_value"><div class="message-tags"><span>lm-input</span></div><div class="message-text">what is this<div class="modality-in-text"><details class="pyglove custom-modality"><summary><div class="summary-name">image<span class="tooltip">metadata.image</span></div><div class="summary-title">CustomModality(...)</div></summary><div class="complex-value custom-modality"><details open class="pyglove str"><summary><div class="summary-name">content<span class="tooltip">metadata.image.content</span></div><div class="summary-title">str</div></summary><span class="simple-value str">&#x27;bird&#x27;</span></details></div></details></div></div></div></details>
        """
    )

  def test_html_ai_message(self):
    image = CustomModality('foo')
    user_message = message.UserMessage(
        'What is in this image? <<[[image]]>> this is a test',
        metadata=dict(image=image),
        source=message.UserMessage('User input'),
        tags=['lm-input']
    )
    ai_message = message.AIMessage(
        'My name is Gemini',
        metadata=dict(
            result=pg.Dict(x=1, y=2, z=pg.Dict(a=[12, 323])),
            usage=language_model.LMSamplingUsage(10, 2, 12)
        ),
        tags=['lm-response', 'lm-output'],
        source=user_message,
    )
    self.assert_html_content(
        ai_message.to_html(enable_summary_tooltip=False),
        """
        <details open class="pyglove ai-message lf-message"><summary><div class="summary-title lf-message">AIMessage(...)</div></summary><div class="complex_value"><div class="message-tags"><span>lm-response</span><span>lm-output</span></div><div class="message-text">My name is Gemini</div><div class="message-result"><details open class="pyglove dict"><summary><div class="summary-name">result<span class="tooltip">metadata.result</span></div><div class="summary-title">Dict(...)</div></summary><div class="complex-value dict"><details open class="pyglove int"><summary><div class="summary-name">x<span class="tooltip">metadata.result.x</span></div><div class="summary-title">int</div></summary><span class="simple-value int">1</span></details><details open class="pyglove int"><summary><div class="summary-name">y<span class="tooltip">metadata.result.y</span></div><div class="summary-title">int</div></summary><span class="simple-value int">2</span></details><details class="pyglove dict"><summary><div class="summary-name">z<span class="tooltip">metadata.result.z</span></div><div class="summary-title">Dict(...)</div></summary><div class="complex-value dict"><details class="pyglove list"><summary><div class="summary-name">a<span class="tooltip">metadata.result.z.a</span></div><div class="summary-title">List(...)</div></summary><div class="complex-value list"><table><tr><td><span class="object-key int">0</span><span class="tooltip">metadata.result.z.a[0]</span></td><td><span class="simple-value int">12</span></td></tr><tr><td><span class="object-key int">1</span><span class="tooltip">metadata.result.z.a[1]</span></td><td><span class="simple-value int">323</span></td></tr></table></div></details></div></details></div></details></div><div class="message-usage"><details open class="pyglove lm-sampling-usage"><summary><div class="summary-name">llm usage<span class="tooltip">metadata.usage</span></div><div class="summary-title">LMSamplingUsage(...)</div></summary><div class="complex-value lm-sampling-usage"><table><tr><td><span class="object-key str">prompt_tokens</span><span class="tooltip">metadata.usage.prompt_tokens</span></td><td><span class="simple-value int">10</span></td></tr><tr><td><span class="object-key str">completion_tokens</span><span class="tooltip">metadata.usage.completion_tokens</span></td><td><span class="simple-value int">2</span></td></tr><tr><td><span class="object-key str">total_tokens</span><span class="tooltip">metadata.usage.total_tokens</span></td><td><span class="simple-value int">12</span></td></tr><tr><td><span class="object-key str">num_requests</span><span class="tooltip">metadata.usage.num_requests</span></td><td><span class="simple-value int">1</span></td></tr><tr><td><span class="object-key str">estimated_cost</span><span class="tooltip">metadata.usage.estimated_cost</span></td><td><span class="simple-value none-type">None</span></td></tr><tr><td><span class="object-key str">retry_stats</span><span class="tooltip">metadata.usage.retry_stats</span></td><td><details class="pyglove retry-stats"><summary><div class="summary-title">RetryStats(...)</div></summary><div class="complex-value retry-stats"><table><tr><td><span class="object-key str">num_occurences</span><span class="tooltip">metadata.usage.retry_stats.num_occurences</span></td><td><span class="simple-value int">0</span></td></tr><tr><td><span class="object-key str">total_wait_interval</span><span class="tooltip">metadata.usage.retry_stats.total_wait_interval</span></td><td><span class="simple-value float">0.0</span></td></tr><tr><td><span class="object-key str">total_call_interval</span><span class="tooltip">metadata.usage.retry_stats.total_call_interval</span></td><td><span class="simple-value float">0.0</span></td></tr><tr><td><span class="object-key str">errors</span><span class="tooltip">metadata.usage.retry_stats.errors</span></td><td><details class="pyglove dict"><summary><div class="summary-title">Dict(...)</div></summary><div class="complex-value dict"><span class="empty-container"></span></div></details></td></tr></table></div></details></td></tr><tr><td><span class="object-key str">completion_tokens_details</span><span class="tooltip">metadata.usage.completion_tokens_details</span></td><td><span class="simple-value none-type">None</span></td></tr></table></div></details></div><div class="message-metadata"><details open class="pyglove dict message-metadata"><summary><div class="summary-name message-metadata">metadata<span class="tooltip message-metadata">metadata</span></div><div class="summary-title message-metadata">Dict(...)</div></summary><div class="complex-value dict"><span class="empty-container"></span></div></details></div><details open class="pyglove user-message lf-message"><summary><div class="summary-name lf-message">source<span class="tooltip lf-message">source</span></div><div class="summary-title lf-message">UserMessage(...)</div></summary><div class="complex_value"><div class="message-tags"><span>lm-input</span></div><div class="message-text">What is in this image?<div class="modality-in-text"><details class="pyglove custom-modality"><summary><div class="summary-name">image<span class="tooltip">source.metadata.image</span></div><div class="summary-title">CustomModality(...)</div></summary><div class="complex-value custom-modality"><details open class="pyglove str"><summary><div class="summary-name">content<span class="tooltip">source.metadata.image.content</span></div><div class="summary-title">str</div></summary><span class="simple-value str">&#x27;foo&#x27;</span></details></div></details></div>this is a test</div><div class="message-metadata"><details open class="pyglove dict message-metadata"><summary><div class="summary-name message-metadata">metadata<span class="tooltip message-metadata">source.metadata</span></div><div class="summary-title message-metadata">Dict(...)</div></summary><div class="complex-value dict"><details class="pyglove custom-modality"><summary><div class="summary-name">image<span class="tooltip">source.metadata.image</span></div><div class="summary-title">CustomModality(...)</div></summary><div class="complex-value custom-modality"><details open class="pyglove str"><summary><div class="summary-name">content<span class="tooltip">source.metadata.image.content</span></div><div class="summary-title">str</div></summary><span class="simple-value str">&#x27;foo&#x27;</span></details></div></details></div></details></div></div></details></div></details>
        """
    )
    self.assert_html_content(
        ai_message.to_html(
            key_style='label',
            enable_summary_tooltip=False,
            extra_flags=dict(
                collapse_modalities_in_text=False,
                collapse_llm_usage=True,
                collapse_message_result_level=0,
                collapse_message_metadata_level=0,
                collapse_source_message_level=0,
                source_tag=None,
            ),
        ),
        """
        <details open class="pyglove ai-message lf-message"><summary><div class="summary-title lf-message">AIMessage(...)</div></summary><div class="complex_value"><div class="message-tags"><span>lm-response</span><span>lm-output</span></div><div class="message-text">My name is Gemini</div><div class="message-result"><details class="pyglove dict"><summary><div class="summary-name">result<span class="tooltip">metadata.result</span></div><div class="summary-title">Dict(...)</div></summary><div class="complex-value dict"><table><tr><td><span class="object-key str">x</span><span class="tooltip">metadata.result.x</span></td><td><span class="simple-value int">1</span></td></tr><tr><td><span class="object-key str">y</span><span class="tooltip">metadata.result.y</span></td><td><span class="simple-value int">2</span></td></tr><tr><td><span class="object-key str">z</span><span class="tooltip">metadata.result.z</span></td><td><details class="pyglove dict"><summary><div class="summary-title">Dict(...)</div></summary><div class="complex-value dict"><table><tr><td><span class="object-key str">a</span><span class="tooltip">metadata.result.z.a</span></td><td><details class="pyglove list"><summary><div class="summary-title">List(...)</div></summary><div class="complex-value list"><table><tr><td><span class="object-key int">0</span><span class="tooltip">metadata.result.z.a[0]</span></td><td><span class="simple-value int">12</span></td></tr><tr><td><span class="object-key int">1</span><span class="tooltip">metadata.result.z.a[1]</span></td><td><span class="simple-value int">323</span></td></tr></table></div></details></td></tr></table></div></details></td></tr></table></div></details></div><div class="message-usage"><details class="pyglove lm-sampling-usage"><summary><div class="summary-name">llm usage<span class="tooltip">metadata.usage</span></div><div class="summary-title">LMSamplingUsage(...)</div></summary><div class="complex-value lm-sampling-usage"><table><tr><td><span class="object-key str">prompt_tokens</span><span class="tooltip">metadata.usage.prompt_tokens</span></td><td><span class="simple-value int">10</span></td></tr><tr><td><span class="object-key str">completion_tokens</span><span class="tooltip">metadata.usage.completion_tokens</span></td><td><span class="simple-value int">2</span></td></tr><tr><td><span class="object-key str">total_tokens</span><span class="tooltip">metadata.usage.total_tokens</span></td><td><span class="simple-value int">12</span></td></tr><tr><td><span class="object-key str">num_requests</span><span class="tooltip">metadata.usage.num_requests</span></td><td><span class="simple-value int">1</span></td></tr><tr><td><span class="object-key str">estimated_cost</span><span class="tooltip">metadata.usage.estimated_cost</span></td><td><span class="simple-value none-type">None</span></td></tr><tr><td><span class="object-key str">retry_stats</span><span class="tooltip">metadata.usage.retry_stats</span></td><td><details class="pyglove retry-stats"><summary><div class="summary-title">RetryStats(...)</div></summary><div class="complex-value retry-stats"><table><tr><td><span class="object-key str">num_occurences</span><span class="tooltip">metadata.usage.retry_stats.num_occurences</span></td><td><span class="simple-value int">0</span></td></tr><tr><td><span class="object-key str">total_wait_interval</span><span class="tooltip">metadata.usage.retry_stats.total_wait_interval</span></td><td><span class="simple-value float">0.0</span></td></tr><tr><td><span class="object-key str">total_call_interval</span><span class="tooltip">metadata.usage.retry_stats.total_call_interval</span></td><td><span class="simple-value float">0.0</span></td></tr><tr><td><span class="object-key str">errors</span><span class="tooltip">metadata.usage.retry_stats.errors</span></td><td><details class="pyglove dict"><summary><div class="summary-title">Dict(...)</div></summary><div class="complex-value dict"><span class="empty-container"></span></div></details></td></tr></table></div></details></td></tr><tr><td><span class="object-key str">completion_tokens_details</span><span class="tooltip">metadata.usage.completion_tokens_details</span></td><td><span class="simple-value none-type">None</span></td></tr></table></div></details></div><div class="message-metadata"><details class="pyglove dict message-metadata"><summary><div class="summary-name message-metadata">metadata<span class="tooltip message-metadata">metadata</span></div><div class="summary-title message-metadata">Dict(...)</div></summary><div class="complex-value dict"><span class="empty-container"></span></div></details></div><details class="pyglove user-message lf-message"><summary><div class="summary-name lf-message">source<span class="tooltip lf-message">source</span></div><div class="summary-title lf-message">UserMessage(...)</div></summary><div class="complex_value"><div class="message-tags"><span>lm-input</span></div><div class="message-text">What is in this image?<div class="modality-in-text"><details open class="pyglove custom-modality"><summary><div class="summary-name">image<span class="tooltip">source.metadata.image</span></div><div class="summary-title">CustomModality(...)</div></summary><div class="complex-value custom-modality"><table><tr><td><span class="object-key str">content</span><span class="tooltip">source.metadata.image.content</span></td><td><span class="simple-value str">&#x27;foo&#x27;</span></td></tr></table></div></details></div>this is a test</div><div class="message-metadata"><details class="pyglove dict message-metadata"><summary><div class="summary-name message-metadata">metadata<span class="tooltip message-metadata">source.metadata</span></div><div class="summary-title message-metadata">Dict(...)</div></summary><div class="complex-value dict"><table><tr><td><span class="object-key str">image</span><span class="tooltip">source.metadata.image</span></td><td><details class="pyglove custom-modality"><summary><div class="summary-title">CustomModality(...)</div></summary><div class="complex-value custom-modality"><table><tr><td><span class="object-key str">content</span><span class="tooltip">source.metadata.image.content</span></td><td><span class="simple-value str">&#x27;foo&#x27;</span></td></tr></table></div></details></td></tr></table></div></details></div><details class="pyglove user-message lf-message"><summary><div class="summary-name lf-message">source<span class="tooltip lf-message">source.source</span></div><div class="summary-title lf-message">UserMessage(...)</div></summary><div class="complex_value"><div class="message-tags"></div><div class="message-text">User input</div><div class="message-metadata"><details class="pyglove dict message-metadata"><summary><div class="summary-name message-metadata">metadata<span class="tooltip message-metadata">source.source.metadata</span></div><div class="summary-title message-metadata">Dict(...)</div></summary><div class="complex-value dict"><span class="empty-container"></span></div></details></div></div></details></div></details></div></details>
        """
    )
    self.assert_html_content(
        ai_message.to_html(
            key_style='label',
            enable_summary_tooltip=False,
            extra_flags=dict(
                collapse_modalities_in_text=True,
                collapse_llm_usage=False,
                collapse_message_result_level=1,
                collapse_message_metadata_level=1,
                collapse_source_message_level=2,
                source_tag=None,
            ),
        ),
        """
        <details open class="pyglove ai-message lf-message"><summary><div class="summary-title lf-message">AIMessage(...)</div></summary><div class="complex_value"><div class="message-tags"><span>lm-response</span><span>lm-output</span></div><div class="message-text">My name is Gemini</div><div class="message-result"><details open class="pyglove dict"><summary><div class="summary-name">result<span class="tooltip">metadata.result</span></div><div class="summary-title">Dict(...)</div></summary><div class="complex-value dict"><table><tr><td><span class="object-key str">x</span><span class="tooltip">metadata.result.x</span></td><td><span class="simple-value int">1</span></td></tr><tr><td><span class="object-key str">y</span><span class="tooltip">metadata.result.y</span></td><td><span class="simple-value int">2</span></td></tr><tr><td><span class="object-key str">z</span><span class="tooltip">metadata.result.z</span></td><td><details class="pyglove dict"><summary><div class="summary-title">Dict(...)</div></summary><div class="complex-value dict"><table><tr><td><span class="object-key str">a</span><span class="tooltip">metadata.result.z.a</span></td><td><details class="pyglove list"><summary><div class="summary-title">List(...)</div></summary><div class="complex-value list"><table><tr><td><span class="object-key int">0</span><span class="tooltip">metadata.result.z.a[0]</span></td><td><span class="simple-value int">12</span></td></tr><tr><td><span class="object-key int">1</span><span class="tooltip">metadata.result.z.a[1]</span></td><td><span class="simple-value int">323</span></td></tr></table></div></details></td></tr></table></div></details></td></tr></table></div></details></div><div class="message-usage"><details open class="pyglove lm-sampling-usage"><summary><div class="summary-name">llm usage<span class="tooltip">metadata.usage</span></div><div class="summary-title">LMSamplingUsage(...)</div></summary><div class="complex-value lm-sampling-usage"><table><tr><td><span class="object-key str">prompt_tokens</span><span class="tooltip">metadata.usage.prompt_tokens</span></td><td><span class="simple-value int">10</span></td></tr><tr><td><span class="object-key str">completion_tokens</span><span class="tooltip">metadata.usage.completion_tokens</span></td><td><span class="simple-value int">2</span></td></tr><tr><td><span class="object-key str">total_tokens</span><span class="tooltip">metadata.usage.total_tokens</span></td><td><span class="simple-value int">12</span></td></tr><tr><td><span class="object-key str">num_requests</span><span class="tooltip">metadata.usage.num_requests</span></td><td><span class="simple-value int">1</span></td></tr><tr><td><span class="object-key str">estimated_cost</span><span class="tooltip">metadata.usage.estimated_cost</span></td><td><span class="simple-value none-type">None</span></td></tr><tr><td><span class="object-key str">retry_stats</span><span class="tooltip">metadata.usage.retry_stats</span></td><td><details class="pyglove retry-stats"><summary><div class="summary-title">RetryStats(...)</div></summary><div class="complex-value retry-stats"><table><tr><td><span class="object-key str">num_occurences</span><span class="tooltip">metadata.usage.retry_stats.num_occurences</span></td><td><span class="simple-value int">0</span></td></tr><tr><td><span class="object-key str">total_wait_interval</span><span class="tooltip">metadata.usage.retry_stats.total_wait_interval</span></td><td><span class="simple-value float">0.0</span></td></tr><tr><td><span class="object-key str">total_call_interval</span><span class="tooltip">metadata.usage.retry_stats.total_call_interval</span></td><td><span class="simple-value float">0.0</span></td></tr><tr><td><span class="object-key str">errors</span><span class="tooltip">metadata.usage.retry_stats.errors</span></td><td><details class="pyglove dict"><summary><div class="summary-title">Dict(...)</div></summary><div class="complex-value dict"><span class="empty-container"></span></div></details></td></tr></table></div></details></td></tr><tr><td><span class="object-key str">completion_tokens_details</span><span class="tooltip">metadata.usage.completion_tokens_details</span></td><td><span class="simple-value none-type">None</span></td></tr></table></div></details></div><div class="message-metadata"><details open class="pyglove dict message-metadata"><summary><div class="summary-name message-metadata">metadata<span class="tooltip message-metadata">metadata</span></div><div class="summary-title message-metadata">Dict(...)</div></summary><div class="complex-value dict"><span class="empty-container"></span></div></details></div><details open class="pyglove user-message lf-message"><summary><div class="summary-name lf-message">source<span class="tooltip lf-message">source</span></div><div class="summary-title lf-message">UserMessage(...)</div></summary><div class="complex_value"><div class="message-tags"><span>lm-input</span></div><div class="message-text">What is in this image?<div class="modality-in-text"><details class="pyglove custom-modality"><summary><div class="summary-name">image<span class="tooltip">source.metadata.image</span></div><div class="summary-title">CustomModality(...)</div></summary><div class="complex-value custom-modality"><table><tr><td><span class="object-key str">content</span><span class="tooltip">source.metadata.image.content</span></td><td><span class="simple-value str">&#x27;foo&#x27;</span></td></tr></table></div></details></div>this is a test</div><div class="message-metadata"><details open class="pyglove dict message-metadata"><summary><div class="summary-name message-metadata">metadata<span class="tooltip message-metadata">source.metadata</span></div><div class="summary-title message-metadata">Dict(...)</div></summary><div class="complex-value dict"><table><tr><td><span class="object-key str">image</span><span class="tooltip">source.metadata.image</span></td><td><details class="pyglove custom-modality"><summary><div class="summary-title">CustomModality(...)</div></summary><div class="complex-value custom-modality"><table><tr><td><span class="object-key str">content</span><span class="tooltip">source.metadata.image.content</span></td><td><span class="simple-value str">&#x27;foo&#x27;</span></td></tr></table></div></details></td></tr></table></div></details></div><details open class="pyglove user-message lf-message"><summary><div class="summary-name lf-message">source<span class="tooltip lf-message">source.source</span></div><div class="summary-title lf-message">UserMessage(...)</div></summary><div class="complex_value"><div class="message-tags"></div><div class="message-text">User input</div><div class="message-metadata"><details open class="pyglove dict message-metadata"><summary><div class="summary-name message-metadata">metadata<span class="tooltip message-metadata">source.source.metadata</span></div><div class="summary-title message-metadata">Dict(...)</div></summary><div class="complex-value dict"><span class="empty-container"></span></div></details></div></div></details></div></details></div></details>
        """
    )

  def test_from_chunks_with_empty_str(self):
    chunks = ['Hello', '', 'World']
    msg = message.AIMessage.from_chunks(chunks)
    self.assertEqual(msg.text, 'Hello World')

    chunks = ['', 'hello']
    msg = message.AIMessage.from_chunks(chunks)
    self.assertEqual(msg.text, 'hello')


class MessageConverterTest(unittest.TestCase):

  def test_basics(self):

    class IntConverter(message.MessageConverter):
      OUTPUT_TYPE = int

    class TestConverter(IntConverter):  # pylint: disable=unused-variable
      FORMAT_ID = 'test_format1'

      def to_value(self, m: message.Message) -> int:
        return int(m.text)

      def from_value(self, value: int) -> message.Message:
        return message.UserMessage(str(value))

    class TestConverter2(IntConverter):  # pylint: disable=unused-variable
      FORMAT_ID = 'test_format2'

      def to_value(self, m: message.Message) -> int:
        return int(m.text) + 1

      def from_value(self, value: int) -> message.Message:
        return message.UserMessage(str(value - 1))

    class TestConverter3(message.MessageConverter):  # pylint: disable=unused-variable
      FORMAT_ID = 'test_format3'
      OUTPUT_TYPE = tuple

      def to_value(self, m: message.Message) -> tuple[int, ...]:
        return tuple(int(x) for x in m.text.split(','))

      def from_value(self, value: tuple[int, ...]) -> message.Message:
        return message.UserMessage(','.join(str(x) for x in value))

    self.assertIn('test_format1', message.Message.convertible_formats())
    self.assertIn('test_format2', message.Message.convertible_formats())
    self.assertIn('test_format3', message.Message.convertible_formats())

    self.assertIn(int, message.Message.convertible_types())
    self.assertIn(tuple, message.Message.convertible_types())
    self.assertEqual(
        message.Message.from_value(1, format='test_format1'),
        message.UserMessage('1')
    )
    self.assertEqual(
        message.UserMessage('1').as_format('test_format1'),
        1
    )
    self.assertEqual(
        message.Message.from_value(1, format='test_format2'),
        message.UserMessage('0')
    )
    self.assertEqual(
        message.UserMessage('1').as_format('test_format2'),
        2
    )
    with self.assertRaisesRegex(ValueError, 'Unsupported format: .*'):
      message.UserMessage('1').as_format('test4')

    with self.assertRaisesRegex(TypeError, 'Cannot convert Message to .*'):
      message.UserMessage('1').as_format(float)

    with self.assertRaisesRegex(
        TypeError, 'More than one converters found for output type .*'
    ):
      message.UserMessage('1').as_format(int)
    self.assertEqual(
        message.UserMessage('1,2,3').as_format('test_format3'),
        (1, 2, 3)
    )
    self.assertEqual(
        message.UserMessage('1,2,3').as_format(tuple),
        (1, 2, 3)
    )
    self.assertEqual(
        message.Message.from_value((1, 2, 3)),
        message.UserMessage('1,2,3')
    )

  def test_get_role(self):
    self.assertEqual(
        message.MessageConverter.get_role(message.SystemMessage('hi')),
        'system',
    )
    self.assertEqual(
        message.MessageConverter.get_role(message.UserMessage('hi')),
        'user',
    )
    self.assertEqual(
        message.MessageConverter.get_role(message.AIMessage('hi')),
        'assistant',
    )
    with self.assertRaisesRegex(ValueError, 'Unsupported message type: .*'):
      message.MessageConverter.get_role(message.MemoryRecord('hi'))

  def test_get_message_cls(self):
    self.assertEqual(
        message.MessageConverter.get_message_cls('system'),
        message.SystemMessage,
    )
    self.assertEqual(
        message.MessageConverter.get_message_cls('user'),
        message.UserMessage,
    )
    self.assertEqual(
        message.MessageConverter.get_message_cls('assistant'),
        message.AIMessage,
    )
    with self.assertRaisesRegex(ValueError, 'Unsupported role: .*'):
      message.MessageConverter.get_message_cls('foo')

  def test_safe_read(self):
    self.assertEqual(
        message.MessageConverter._safe_read({'a': 1}, 'a'),
        1,
    )
    self.assertEqual(
        message.MessageConverter._safe_read({'a': 1}, 'a', default=2),
        1,
    )
    self.assertEqual(
        message.MessageConverter._safe_read({'a': 1}, 'b', default=2),
        2,
    )
    with self.assertRaisesRegex(ValueError, 'Invalid data type: .*'):
      message.MessageConverter._safe_read(1, 'a')
    with self.assertRaisesRegex(ValueError, 'Missing key .*'):
      message.MessageConverter._safe_read({'a': 1}, 'b')

if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Evaluation (v1) for Langfun agentic actions."""

import io
import os
from typing import Annotated, Any

import langfun.core as lf
from langfun.core import eval as lf_eval
from langfun.core.agentic import action as action_lib
import pyglove as pg


class ActionEval(lf.eval.v2.Evaluation):
  """Agent evaluation."""

  action_args: Annotated[
      dict[str, Any],
      'Arguments to call the action.'
  ] = {}

  def process(self, example: lf.eval.v2.Example) -> tuple[str, dict[str, Any]]:
    example_input = example.input
    action = example_input.action

    # We explicitly create a session here to use a custom session ID.
    with action_lib.Session(
        id=f'{self.id}#example-{example.id}', verbose=True
    ) as session:

      # NOTE(daiyip): Setting session as metadata before action execution, so we
      # could use `Evaluation.state.in_progress_examples` to access the session
      # for status reporting from other threads.
      example.metadata['session'] = session
      action(session=session, **self.action_args)

    return session.final_result, dict(session=session)

#
# TODO(daiyip): Remove V1 once V2 is fully launched.
#


@pg.functor()
def _dummy_schema():
  return int


class ExampleView(pg.Object):
  id: int
  input: Any
  output: Any
  error: str | None = None


class ActionEvalV1(lf_eval.Matching):
  """Base class for action evaluations.

  The input function should returns a list of pg.Dict, with `action` and
  `groundtruth` fields.
  """
  # We override the schema and prompt to dummy values since they are not used.
  schema_fn = _dummy_schema()
  prompt = '<unused>'

  def process(self, example: pg.Dict, **kwargs):
    action = example.action
    with action_lib.Session(
        id=str(getattr(example, 'id', '<empty>'))) as session:
      action(session=session, lm=self.lm, **kwargs)
    return session.as_message()

  def answer(self, output: Any, example: pg.Dict) -> Any:
    return output

  def groundtruth(self, example: Any) -> Any:
    return example.groundtruth

  def audit(
      self,
      example_idx: int,
      example: Any,
      message: lf.Message | None,
      error: Exception | None = None,
      dryrun: bool = False,
  ):
    super().audit(example_idx, example, message, error, dryrun)
    # Write each example to HTML.
    if not dryrun and self.dir:
      def _save_html():
        ExampleView(
            example_idx,
            example,
            None if message is None else message.result,
            error
        ).to_html(
            collapse_level=None,
            enable_summary_tooltip=False,
        ).save(
            os.path.join(self.dir, f'example_{example_idx}.html')
        )
      # Write HTML in a separate thread to avoid blocking the main thread.
      lf.concurrent.get_executor(
          'background_eval_io', max_workers=16
      ).submit(_save_html)

  def _render_mismatches(self, s: io.StringIO) -> None:
    s.write('<h2> Mismatches (Incorrect) </h2>')
    first_url = None
    mismatched_ids = sorted([
        example_idx for example_idx, *_ in self.mismatches
    ])
    for example_idx in mismatched_ids:
      url = os.path.join(self.dir, f'example_{example_idx}.html')
      if first_url is None:
        first_url = url
      s.write(
          f'<a href="{url}" style="margin-right: 10px" target="example_view">'
          f'{example_idx}</a> '
      )
    if first_url:
      s.write(
          '<iframe style="border:0;width:100%;height:100%" name="example_view"'
          f'src="{first_url}" title="Example View"></iframe>'
      )
    else:
      s.write('No mismatches found.')

  def _render_matches(self, s: io.StringIO) -> None:
    s.write('<h2> Matches (correct) </h2>')
    first_url = None
    matched_ids = sorted([
        example_idx for example_idx, *_ in self.matches
    ])
    for example_idx in matched_ids:
      url = os.path.join(self.dir, f'example_{example_idx}.html')
      if first_url is None:
        first_url = url
      s.write(
          f'<a href="{url}" style="margin-right: 10px">{example_idx}</a> '
      )
    if first_url:
      s.write(
          '<iframe style="border:0;width:100%;height:100%" name="example_view"'
          f'src="{first_url}" title="Example View"></iframe>'
      )
    else:
      s.write('No matches found.')



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Language models from llama.cpp."""

from typing import Annotated
from langfun.core.llms import openai_compatible
import pyglove as pg


@pg.use_init_args(['url', 'model'])
@pg.members([('api_endpoint', pg.typing.Str().freeze(''))])
class LlamaCppRemote(openai_compatible.OpenAICompatible):
  """The remote LLaMA C++ model.

  The Remote LLaMA C++ models can be launched via
  https://github.com/ggerganov/llama.cpp/tree/master/examples/server
  """
  url: Annotated[
      str,
      'The URL of the LLaMA C++ server.',
  ]

  model: Annotated[
      str,
      'The name of the model to use.',
  ] = ''

  @property
  def api_endpoint(self) -> str:
    return self.url + '/completion'

  @property
  def model_id(self) -> str:
    """Returns a string to identify the model."""
    return f'LLaMAC++({self.model or ""})'




# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for code correction."""

import inspect
import unittest

from langfun.core.coding.python import correction
from langfun.core.llms import fake
import pyglove as pg


class RunWithCorrectionTest(unittest.TestCase):

  def test_run_with_correction(self):
    result = correction.run_with_correction(
        inspect.cleandoc("""
            x = 1,
            y = x + 2
             z = x + y
            """),
        lm=fake.StaticSequence([
            inspect.cleandoc("""
                CorrectedCode(
                    corrected_code='x = 1,\\ny = x + 2\\nz = x + y',
                )
                """),
            inspect.cleandoc("""
                CorrectedCode(
                    corrected_code='x = 1\\ny = x + 2\\nz = x + y',
                )
                """),
        ]),
    )
    self.assertEqual(result, 4)

  def test_run_with_correction_upon_custom_validation(self):

    class Foo(pg.Object):
      x: int

      def __validate__(self):
        if self.x > 1:
          raise ValueError('value should be less or equal than 1.')
        if self.x < 0:
          self.rebind(x=0, skip_notification=True)

    result = correction.run_with_correction(
        inspect.cleandoc("""
            Foo(x=2)
            """),
        global_vars=dict(Foo=Foo),
        lm=fake.StaticSequence([
            inspect.cleandoc("""
                CorrectedCode(
                    corrected_code='Foo(x=-1)',
                )
                """),
        ]),
    )
    self.assertEqual(result, Foo(0))

  def test_run_without_correction(self):
    result = correction.run_with_correction(
        inspect.cleandoc("""
            x = 1
            y = x + 2
            z = x + y
            """),
        max_attempts=0,
    )
    self.assertEqual(result, 4)
    with self.assertRaises(pg.coding.CodeError):
      correction.run_with_correction(
          inspect.cleandoc("""
            x = 1,
            y = x + 2
            z = x + y
            """),
          max_attempts=0,
      )


class CorrectTest(unittest.TestCase):

  def test_correct(self):
    corrected = correction.correct(
        inspect.cleandoc("""
            x = 1,
            y = x + 2
            z = x + y
            """),
        lm=fake.StaticSequence([
            inspect.cleandoc("""
                CorrectedCode(
                    corrected_code='x = 1,\\ny = x + 2\\nz = x + y',
                )
                """),
            inspect.cleandoc("""
                CorrectedCode(
                    corrected_code='x = 1\\ny = x + 2\\nz = x + y',
                )
                """),
        ]),
    )
    self.assertEqual(corrected, 'x = 1\ny = x + 2\nz = x + y')

  def test_correct_good_code(self):
    self.assertEqual(
        correction.correct('x + y + z', global_vars=dict(x=0, y=1, z=2)),
        'x + y + z',
    )

  def test_correct_reaching_limit(self):
    with self.assertRaisesRegex(
        pg.coding.CodeError, 'Cannot correct code after 1 attempts'
    ):
      correction.correct(
          inspect.cleandoc("""
              x = 1,
              y = x + 2
              z = x + y
              """),
          (
              'IndentationError: unexpected indent (<unknown>, line 3)\n'
              '  z = x + y'
          ),
          lm=fake.StaticSequence([
              inspect.cleandoc("""
                  CorrectedCode(
                      corrected_code='x = 1,\\ny = x + 2\\nz = x + y',
                  )
                  """),
          ]),
          max_attempts=1,
      )


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import os
import tempfile
import unittest

from langfun.core.coding.python import sandboxing
import pyglove as pg


class MultiProcessingSandboxTest(unittest.TestCase):
  """Tests for MultiProcessingSandbox."""

  def test_basics(self):
    with tempfile.TemporaryDirectory() as dir1:
      input_file = os.path.join(dir1, 'test.json')
      pg.save(pg.Dict(x=1, y=2), input_file)
      with sandboxing.MultiProcessingSandbox() as sandbox:
        self.assertIsNotNone(sandbox.working_dir)
        sandbox_input_file = sandbox.upload(input_file)
        code = (
            f"""
            import pyglove as pg
            print(pg.load({input_file!r}).x)
            """
        )
        self.assertEqual(
            sandbox.normalize_code(code),
            inspect.cleandoc(
                f"""
                import pyglove as pg
                print(pg.load({sandbox_input_file!r}).x)
                """
            ),
        )
        self.assertEqual(sandbox.run(code).stdout, '1\n')

  def test_bad_code(self):
    with sandboxing.MultiProcessingSandbox() as sandbox:
      with self.assertRaisesRegex(pg.coding.CodeError, '.* is not defined'):
        sandbox.run('print(x)')

  def test_bad_usages(self):
    sandbox = sandboxing.MultiProcessingSandbox()
    with self.assertRaisesRegex(ValueError, 'Sandbox is not set up.'):
      sandbox.upload('abc.txt')


if __name__ == '__main__':
  unittest.main()
